#!/usr/bin/env python
# -*- coding: ascii -*-


r"""
CoolProp has a set of incompressible fluids
EC_Incomp_Fluid is a wrapper of those fluids using Engineering Units.

EngCoolProp uses units of primarily inch, lbm, lbf, sec, BTU (some use of ft and hour).::

    #:c     The following are the default units for each property of Incompressible Fluids.
    #:c 
    #:c     T = Temperature = degR
    #:c     P = Pressure = psia
    #:c     D = Density = lbm/cu ft
    #:c     rho = Density = lbm/cu inch
    #:c     E = Internal Energy = BTU/lbm
    #:c     H = Enthalpy = BTU/lbm
    #:c     S = Entropy = BTU/lbm degR
    #:c     Cp = Heat Capacity (const. P) = BTU/lbm degR
    #:c     V = Viscosity = 1.0E5 * lbm/ft-sec
    #:c     C = Thermal Conductivity = BTU/ft-hr-R

"""
import traceback
from engcoolprop.ec_fluid import (CPeng_fromSI ,  CondEng_fromSI ,   DSI_fromEng ,  
                                  Deng_fromSI ,   PSI_fromEng, Peng_fromSI,
                                  PropsSI ,   SSI_fromEng ,  Seng_fromSI ,  TSI_fromEng ,  
                                  Teng_fromSI ,  UHSI_fromEng ,  UHeng_fromSI ,   Veng_fromSI )
from CoolProp.CoolProp import PropsSI
import CoolProp.CoolProp as CP
from engcoolprop.safe_get_property import safe_get_INCOMP_prop
from engcoolprop.find_exception_threshold import find_exception_limit

from engcoolprop.InterpProp_scipy import get_density_interpolator
from engcoolprop.iteration_utils import find_T_from_Dterp, find_T_at_P
from engcoolprop.iteration_utils import calc_Tnbp
from engcoolprop.utils import Same_g_len, print_avoid_valerr, incomp_pure_fluidL
from engcoolprop.banner import banner


# create simple look-up that is order-independent for input pairs

call_tuplesD = {}  # key:tuple indep vars (e.g. ("T","P")), value: dependent property (e.g. "S", "H", etc.)
call_tuplesD[('D', 'P')] = {'L', 'T', 'S', 'U', 'H', 'V', 'C'}
call_tuplesD[('H', 'P')] = {'L', 'T', 'S', 'U', 'V', 'C', 'D'}
call_tuplesD[('P', 'D')] = {'L', 'T', 'S', 'U', 'H', 'V', 'C'}
call_tuplesD[('P', 'H')] = {'L', 'T', 'S', 'U', 'V', 'C', 'D'}
call_tuplesD[('P', 'S')] = {'L', 'T', 'U', 'H', 'V', 'C', 'D'}
call_tuplesD[('P', 'T')] = {'L', 'S', 'U', 'H', 'V', 'C', 'D'}
call_tuplesD[('S', 'P')] = {'L', 'T', 'U', 'H', 'V', 'C', 'D'}
call_tuplesD[('T', 'P')] = {'L', 'S', 'U', 'H', 'V', 'C', 'D'}          


# how to make a list of all incompressible fluids in coolprop
# incomp_pure_fluidL = CP.get_global_param_string('incompressible_list_pure').split(',')

class EC_Incomp_Fluid(object):
    
    def __init__(self,symbol="DowQ", T=None ,P=None, Pmax=10000.0,
                 show_warnings=2, auto_fix_value_errors=False):
        """        Init generic Incompressible Fluid

        Args:
            symbol (str): Name and percent mass. (e.g. "Water", "DowQ")
            T (None or float): Temperature degR (None sets T to (Tmin+Tmax)/2)
            P (None or float): Pressure psia. (None set P to Pmax/10)
            Pmax (float): Max expected pressure psia. Defaults to 10000.0.
            show_warnings (int): Sets warning level(0=None, 1=Only serious, 2=all). Defaults to 2.
            auto_fix_value_errors (bool): Action when ValueError occurs. 
                                        True=correct problem with a warning, 
                                        False throws Exception Defaults to False.
        """

        if symbol not in incomp_pure_fluidL:
            raise ValueError( '"%s" is NOT in coolprop incompressible list\n%s'%(symbol, repr(incomp_pure_fluidL) ) )
        
        self.symbol = symbol
        self.Pmax = Pmax # highest pressure considered in any iterative calcs (can still input P > Pmax)
        self.Pmax_si = PSI_fromEng( Pmax )
        if P is None:
            P = int( Pmax / 10 )

        self.show_warnings = show_warnings # some calcs will issue warning statements if show_warnings >0 severe>1
        self.auto_fix_value_errors = auto_fix_value_errors

        if show_warnings>1 and auto_fix_value_errors:
            banner( 'NOTICE: any input violations on limits of T, D, H or S\nwill be automatically corrected (set to min or max).'+\
                   '\nTo change this behavior set "auto_fix_value_errors" to False'+\
                    '\nTo suppress this banner set "show_warnings" to 0 or 1')

        self.fluid = 'INCOMP::%s'%symbol

        # get temperature limits (Note attempt to avoid CoolProp round-off problems)
        self.Tmin_si =   PropsSI('Tmin','T',0,'P',0,'INCOMP::%s'%symbol) * 1.0000000000000002
        self.Tmax_si =   PropsSI('Tmax','T',0,'P',0,'INCOMP::%s'%symbol) * 0.9999999999999999

        # ==== some fluids have min or max point in density curve ( screws up state = f(P, D) ) ====
        if self.symbol == 'PMR':
            self.Tmax_si = 500.0 # degK # has minimum in D=f(T) curve.
            if show_warnings > 0:
                print( 'WARNING: PMR Tmax reduced to 500 degK (900 degR) due to error in CoolProp equation for Density' )
        elif self.symbol == 'NBS':
            self.Tmin_si = 275.0 # degK # has maximum in D=f(T) curve.
            if show_warnings > 0:
                print( 'WARNING: NBS Tmin increased to 275 degK (495 degR) due to error in CoolProp equation for Density' )
        elif self.symbol == 'FoodWater':
            self.Tmin_si =  280.0 # degK # has maximum in D=f(T) curve.
            if show_warnings > 0:
                print( 'WARNING: FoodWater Tmin increased to 280 degK (504 degR) due to error in CoolProp equation for Density' )
        elif self.symbol == 'LiqNa':
            self.Tmax_si = 2455.0 # degK # has maximum in H=f(T,P) curve.
            if show_warnings > 0:
                print( 'WARNING: LiqNa Tmax reduced to 2455 degK (4419 degR) due to error in CoolProp equation for Enthalpy' )



        self.Tmin =  Teng_fromSI( self.Tmin_si  )
        self.Tmax =  Teng_fromSI( self.Tmax_si )

        self.Tmid = (self.Tmin + self.Tmax) / 2.0

        self.Dterp = get_density_interpolator( self )

        self.check_visc_support()

        # Note that Psat_min and Psat_max are calculated in calc_min_max_props
        self.Pmin = self.get_Psat( self.Tmin ) #+ 0.0001

        if T is None:
            T = int((self.Tmin + self.Tmax) / 2.0)
            # else: T = input T
        
        self.P = P
        self.Pinput = P # in setTP, if P is below Psat, P will be increased to Psat

        try:
            Tnbp, error_code = calc_Tnbp( self )
            if not error_code:
                self.Tnbp = Tnbp 
            else:
                self.Tnbp = None 
        except:
            self.Tnbp = None 

        # if input T is in range, use it... otherwise
        if T<self.Tmin:
            if self.auto_fix_value_errors:
                self.T = self.Tmin + 0.001
                print( 'NOTICE: input T too low. Changed from: %g to: %g degR'%(T, self.T) )
                print( "        INCOMP: Tmin=%g, Tmax=%g"%(self.Tmin, self.Tmax) )
            else:
                print_avoid_valerr()
                raise ValueError("Input T too low. T=%g,  Tmin=%g"%(T, self.Tmin))

        elif T>self.Tmax:
            if self.auto_fix_value_errors:                            
                self.T = self.Tmax - 0.001
                print( 'NOTICE: input T too high. Changed from: %g to: %g degR'%(T, self.T) )
                print( "        INCOMP: Tmin=%g, Tmax=%g"%(self.Tmin, self.Tmax) )
            else:
                print_avoid_valerr()
                raise ValueError("Input T too high. T=%g,  Tmax=%g"%(T, self.Tmax))
        else:
            self.T = T

        self.calc_min_max_props()

        # set properties to input T and P
        # print( 'First call to setTP')
        self.setTP(self.T, self.P)
        # print( 'Back from: First call to setTP')
            
            
    def check_visc_support(self):
        try:
            _ =  PropsSI('V','T',self.Tmin_si,'P',self.Pmax_si, self.fluid ) # Vmax at Tmin
            _ =  PropsSI('V','T',self.Tmax_si,'P',self.Pmax_si, self.fluid ) # Vmin at Tmax
            self.visc_is_supported = True
        except:
            self.visc_is_supported = False
            if self.show_warnings:
                print( 'check_visc_support: Visc NOT supported for fluid:', self.symbol)

        # print( 'Found visc_is_supported =', self.visc_is_supported)



    def set_warnings(self, show_warnings=2):
        """Set the show_warnings flag"""
        if show_warnings is True:
            show_warnings = 1
        elif show_warnings is False:
            show_warnings = 0

        self.save_show_warnings = self.show_warnings
        self.show_warnings = show_warnings

    def pause_warnings(self):
        """Pause all warnings until self.show_warnings is changed"""
        self.save_show_warnings = self.show_warnings
        self.show_warnings = 0

    def restore_warnings(self):
        """Restore the self.show_warnings flag to the last pause_warnings condition"""
        try:
            self.show_warnings = self.save_show_warnings
        except:
            pass
    

    def setProps(self, **inpD):
        '''Generic call using any P with supported inputs T,D,S,H::
        
        #: for example
        #: ec.setProps(T=100, P=200)
        #: ec.setProps(D=0.1, P=100)
        #: ec.setProps(P=100, H=20)
        #:
        #: Made so both:
        #: setProps(T=530, P=100)  AND  
        #: setProps(P=100, T=530) will work

        '''
        
        key = tuple( inpD.keys() )
        if key not in call_tuplesD:
            print('ERROR... setProps called with unknown inputs', inpD)
            print('Allowable inputs are: T,P,D,H,S')
            if len(inpD) != 2:
                print('ONLY 2 inputs are allowed.  (e.g. T=530.0,P=100.0)')
            raise ValueError('setProps called with illegal inputs: %s'%str(inpD))
        
        if 'P' not in inpD:
            raise ValueError('"P" must be one of the setProps inputs.')
        
        if len(inpD) != 2:
            raise ValueError('ONLY 2 inputs are allowed.  (e.g. T=530.0,P=100.0), not' + repr(inpD))

        if "T" in inpD:
            self.setTP( ** inpD )
        elif "D" in inpD:
            self.setPD( **inpD )
        elif "H" in inpD:
            self.setPH( **inpD )
        elif "S" in inpD:
            self.setPS( **inpD )
        else:
            raise ValueError( 'Invalid input to setProps ' + repr(inpD) )
        
    def get_D_at_T(self, T): # degR
        Psi = PSI_fromEng( 5000 )
        
        Tsi = TSI_fromEng( T )
        
        D = Deng_fromSI( PropsSI('D', 'T',Tsi,'P',Psi, self.fluid) )
        return D # lbm/cuft


    def get_Psat(self, T): # degR
        """
        Given T in degR (Engineering units) return Psat in psia (if supported)
        NOTE: if not supported, return 0 for Psat
        """
        try:
            Psat = Peng_fromSI( PropsSI('P','T', TSI_fromEng( T ) ,'Q',0, self.fluid) )
        except:
            Psat = 0.0
        return Psat # psia


    def get_Tsat(self, P ): # psia
        """
        Calculate Tsat at given P (psia).
        If P is greater than max Psat, simply return Tmax.
        """
        raise Exception( 'INCOMP Psat data not complete enough for Tsat Calculation')


    def setTP(self,T=530.0,P=1000.0):
        '''Calc props from T and P'''


        if T < self.Tmin:
            if self.auto_fix_value_errors:
                if self.show_warnings and self.Tmin - T > 0.01:
                    print( 'T too low in setTP. Changed T=%g to Tmin=%g'%( T, self.Tmin ))
                T = self.Tmin
            else:
                print_avoid_valerr()
                raise ValueError("Input T too low in setTP. T=%g,  Tmin=%g"%(T, self.Tmin))
        if T > self.Tmax:
            if self.auto_fix_value_errors:
                # if self.show_warnings and T - self.Tmax  > 0.01:
                if self.show_warnings and T > self.Tmax + 0.1 :
                    print( 'T too high in setTP. Changed T=%g to Tmax=%g'%( T, self.Tmax ))
                T = self.Tmax
            else:
                print_avoid_valerr()
                raise ValueError("Input T too high in setTP. T=%g,  Tmin=%g"%(T, self.Tmax))

        self.Pinput = P # save input P in case Psat changes it.

        # === Don't allow Vapor phase... increase P to phase line ===
        # P = max( P,  self.get_Psat( T ))

        self.Psat = self.get_Psat( self.T )

        # if P > self.Pinput:
        if self.Psat > P:
            # print( 'Psat - P =', self.Psat - P)
            if self.auto_fix_value_errors:
                P = self.Psat
                if self.show_warnings:
                    print( 'P too low in setTP. Changed P=%g to Psat=%g'%( self.Pinput, self.Psat ))
            else:
                print_avoid_valerr()
                raise ValueError( 'P too low in setTP. P=%g should be >= Psat=%g'%( self.Pinput, self.Psat ))


        
        Tsi = TSI_fromEng( T )
        Psi = PSI_fromEng( P )
        def get_prop( prop_desc='H' ):
            try:
                prop, good_Psi = safe_get_INCOMP_prop( prop_desc, Psi_val=Psi, ind_name='T', ind_si_val=Tsi, 
                                                       symbol=self.symbol, show_warnings=self.show_warnings>1,
                                                       Pmax=self.Pmax )
                return prop
            except:
                return float('inf')

        self.T = T
        self.P = P

        # self.D = Deng_fromSI( get_prop('D') )
        self.D = self.get_D_at_T( self.T )

        self.rho = self.D / 1728.0

        self.E = UHeng_fromSI( get_prop('U') )
        self.H = UHeng_fromSI( get_prop('H') )
        self.S = Seng_fromSI( get_prop('S') )
        self.Cp = CPeng_fromSI( get_prop('C') )

        # Some INCOMP fluids have no viscosity data
        if self.visc_is_supported:
            self.Visc = Veng_fromSI( get_prop('V') ) * 1.0E5
        else:
            self.Visc = float('inf')

        self.Cond = CondEng_fromSI( get_prop('L') )


        # self.Tsat = self.get_Tsat( self.P )

        # ============== Debug prints ================
        # if self.H > self.Hmax:
        #     print( '...... WARNING: in setTP got H=%g > Hmax=%g ..................'%(self.H, self.Hmax))

        # print( '.......................EXITING setTP..........................')
        
    def constP_newH(self,H):
        '''Calc properties at new H with same P'''

        self.setPH( self.P, H)
        
    def constH_newP(self,P=1000.0):
        '''Calc properties at new P with same H'''

        self.setPH( P, self.H)
        
    def setPH(self,P,H):
        '''Calc properties at P and H'''

        if self.symbol == 'Air':
            raise Exception( 'PH data for Air will not work properly in setPH.' )
  
        if H < self.Hmin:
            if self.auto_fix_value_errors:
                if self.show_warnings:
                    print( 'H too low in setPH. Changed H=%g to H=%g'%( H, self.Hmin ))
                H = self.Hmin
            else:
                print_avoid_valerr()
                raise ValueError("Input H too low in setPH. H=%g,  Hmin=%g"%(H, self.Hmin))
        if H > self.Hmax:
            if self.auto_fix_value_errors:
                if self.show_warnings:
                    print( 'H too high in setPH. Changed H=%g to Hmax=%g'%( H, self.Hmax ))
                H = self.Hmax
            else:
                print_avoid_valerr()
                raise ValueError("Input H too high in setPH. H=%g,  Hmax=%g"%(H, self.Hmax))


        self.Pinput = P # save input P in case Psat changes it.
        # P = self.adjust_P_for_Psat( P, ind_param="H", ind_val=H )
        # if P > self.Pinput:
        #     if self.show_warnings:
        #         print( 'P too low in setPH. Changed P=%g to Psat=%g'%( self.Pinput, P ))
        
        self.P = P 
        self.H = H
        
        Psi = PSI_fromEng( P )
        Hsi = UHSI_fromEng( H )

        # This approach worked for PS, but not PH
        # self.T, err_flag = find_T_at_P(self, P, dep_name='H', dep_val=H, tol=1.0E-12)
        # self.setTP( self.T, self.P)
        # return

        def get_prop( prop_desc='T' ):
            try:
                prop, good_Psi = safe_get_INCOMP_prop( prop_desc, Psi_val=Psi, ind_name='H', ind_si_val=Hsi, 
                                                       symbol=self.symbol, show_warnings=self.show_warnings>1,
                                                       Pmax=self.Pmax )
                return prop
            except:
                return float('inf')

        self.T = Teng_fromSI( get_prop('T') )

        
        self.setTP( self.T, self.P)
        
    def setPS(self,P,S):
        '''Calc properties at P and H'''

        if self.symbol == 'Air':
            raise Exception( 'PS data for Air will not work properly in setPS.' )

        if S < self.Smin:
            if self.auto_fix_value_errors:
                if self.show_warnings:
                    print( 'S too low in setPS. Changed S=%g to S=%g'%( S, self.Smin ))
                S = self.Smin
            else:
                print_avoid_valerr()
                raise ValueError("Input D too low in setPS. S=%g,  Smin=%g"%(S, self.Smin))
            
        if S > self.Smax:
            if self.auto_fix_value_errors:
                if self.show_warnings:
                    print( 'S too high in setPS. Changed S=%g to Smax=%g'%( S, self.Smax ))
                S = self.Smax
            else:
                print_avoid_valerr()
                raise ValueError("Input D too low in setPS. S=%g,  Smin=%g"%(S, self.Smin))

        self.Pinput = P # save input P in case Psat changes it.
        # P = self.adjust_P_for_Psat( P, ind_param="S", ind_val=S )
        # if P > self.Pinput:
        #     if self.show_warnings:
        #         print( 'P too low in setPS. Changed P=%g to Psat=%g'%( self.Pinput, P ))

        self.P = P 
        self.S = S

        self.T, err_flag = find_T_at_P(self, P, dep_name='S', dep_val=S, tol=1.0E-12)
        self.setTP( self.T, self.P)

    def constS_newP(self,P=1000.0):
        '''Calc properties at new P with same S'''

        self.setPS( P, self.S )
        
    def setPD(self,P=1000.0,D=0.01):
        '''
        Calc props from P and D
        NOTE: The pressure has NO EFFECT on calculated temperature for incompressible density.
        '''

        if D < self.Dmin:
            if self.auto_fix_value_errors:
                if self.show_warnings:
                    print( 'D too low in setDP. Changed D=%g to D=%g'%( D, self.Dmin ))
                D = self.Dmin
            else:
                print_avoid_valerr()
                raise ValueError("Input D too low in setPD. D=%g,  Dmin=%g"%(D, self.Dmin))

        if D > self.Dmax:
            if self.auto_fix_value_errors:
                if self.show_warnings:
                    print( 'D too high in setDP. Changed D=%g to Dmax=%g'%( D, self.Dmax ))
                D = self.Dmax
            else:
                print_avoid_valerr()
                raise ValueError("Input D too high in setPD. D=%g,  Dmax=%g"%(D, self.Dmax))

        self.Pinput = P # save input P in case Psat changes it.
        # P = self.adjust_P_for_Psat( P, ind_param="D", ind_val=D )

        if P > self.Pinput:
            if self.show_warnings:
                print( 'P too low in setPD. Changed P=%g to Psat=%g'%( self.Pinput, P ))

        self.P = P 
        self.D = D

        self.T, err_flag = find_T_from_Dterp(self, D)
        if err_flag and self.show_warnings:
            print( 'WARNING: in setPD, find_T_from_Dterp has an error.')

        self.setTP( self.T, self.P )
        
        
    def getStrTransport(self):
        '''create a string from the Transport properties'''
        return  "%s Cp=%6g Visc=%6g ThCond=%6g" %\
        (self.fluid,self.Cp, self.Visc, self.Cond)

    def getStrTPD(self):
        '''create a string from the TPDEHS properties'''
        return  "%s T=%6.1f P=%6.1f D=%.4f E=%6.2f H=%6.2f S=%.3f" %\
        (self.fluid,self.T, self.P, self.D, self.E, self.H, self.S)

    # def getStrTPDphase(self):
    #     '''create a string from the TPDEHS properties'''
    #     return  "%s T=%6.1f P=%6.1f D=%.4f E=%6.2f H=%6.2f S=%.3f" %\
    #     (self.fluid,self.T, self.P, self.D, self.E, self.H, self.S)

    def printTPD(self):
        '''print a string from the TPDEHS properties'''
        print(self.getStrTPD())

    def printTransport(self):
        '''print a string of Transport properties'''
        print(self.getStrTransport())

    def printProps(self):
        '''print a multiline property summary with units'''

        # get formatted floats that are same length to help the look of table
        SGL = Same_g_len(self, ['Cond', "Cp", 'D', 'E', 'H', 'P', 'S', 'T', 'Visc', 'rho', 'Tnbp', 'Psat'] ) # , 'Tsat'


        print("State Point for fluid",self.fluid,"("+ self.symbol +")")
        print("T =%s"%SGL.T," degR,",    '                             Range(%8g - %8g) degR'%(self.Tmin, self.Tmax))

        print("P =%s"%SGL.P," psia",    '                              Range(%8g - %8g) psia'%(self.Pmin, self.Pmax))
        if self.Pinput != self.P:
            print("    Pinput =%8g"%self.Pinput, '(Adjusted due to Psat)')

        print("D =%s"%SGL.D," lbm/cuft",    '                          Range(%8g - %8g) lbm/cuft'%(self.Dmin, self.Dmax))

        print("E =%s"%SGL.E," BTU/lbm",    '                           Range(%8g - %8g) BTU/lbm'%(self.Emin, self.Emax))
        print("H =%s"%SGL.H," BTU/lbm",    '                           Range(%8g - %8g) BTU/lbm'%(self.Hmin, self.Hmax))
        print("S =%s"%SGL.S," BTU/lbm degR",    '                      Range(%.6f - %.6f) BTU/lbm degR'%(self.Smin, self.Smax))
        print("Cp=%s"%SGL.Cp," BTU/lbm degR",   '                      Range(%8g - %8g) BTU/lbm degR'%(self.Cpmin, self.Cpmax))

        if self.Visc < float('inf'):
            print("V =%s"%SGL.Visc," viscosity [1.0E5 * lbm/ft-sec]", '    Range(%8g - %8g)'%(self.Viscmin, self.Viscmax) )
        else:
            print("V =UNDEFINED","viscosity [1.0E5 * lbm/ft-sec]" )

        print("C =%s"%SGL.Cond," thermal conductivity [BTU/ft-hr-R]", 'Range(%8g - %8g)'%(self.Condmin, self.Condmax) )

        if self.Tnbp is not None:
            print("    Tnbp =%s"%SGL.Tnbp," degR,")
        print("    rho  =%s"%SGL.rho," lbm/cuin",   '                   Range(%.6f - %.6f) lbm/cuin'%(self.rho_min, self.rho_max))
        
        if self.psat_is_supported:
            print("    Psat =%s"%SGL.Psat," psia", '                       Range(%g - %g) psia'%(self.Psat_min, self.Psat_max))
            # print("    Tsat =%s"%SGL.Tsat," degR", '                       Range(%g - %g) degR'%(self.Tsat_min, self.Tsat_max))

    def calc_min_max_props(self, do_print=False):
        # print( '.......................Entered calc_min_max_props ..........................')
        T_save = self.T
        P_save = self.P

        self.setTP(self.T, self.P)

        self.Psat_min = self.get_Psat( self.Tmin )
        self.Psat_max = self.get_Psat( self.Tmax )
        if self.Psat_max > 0:
            self.psat_is_supported = True
        else:
            self.psat_is_supported = False

        # self.Tsat_min = self.get_Tsat( self.Pmin )
        # self.Tsat_max = self.get_Tsat( self.Pmax )


        # make a list of the four corners of the T,P space
        tpL = [(self.Tmin, 0), (self.Tmin, self.Pmax), (self.Tmax, 0), (self.Tmax, self.Pmax)]

        resultD = {} # key:min/max name (e.g. Dmin, Cpmax), value:float value

        # set starting values for min/max of these properties
        for prop_name in [ 'D', 'E', 'H', 'S', 'Cp', 'Visc', 'Cond' ]:
            resultD[prop_name+'min'] = float('inf')
            resultD[prop_name+'max'] = float('-inf')

        self.pause_warnings()
        for T,P in tpL:
            self.setTP( T, P )

            for prop_name in [ 'D', 'E', 'H', 'S', 'Cp', 'Visc', 'Cond' ]:
                if getattr(self, prop_name) < resultD[prop_name+'min']:
                    resultD[prop_name+'min'] = getattr(self, prop_name)

                if getattr(self, prop_name) > resultD[prop_name+'max']:
                    resultD[prop_name+'max'] = getattr(self, prop_name)
        self.restore_warnings()

        # set min/max values found in resultD
        for prop_name, value in resultD.items():
            # print( 'setting ', prop_name, value)
            setattr( self, prop_name, value )

        self.rho_min = self.Dmin / 1728.0
        self.rho_max = self.Dmax / 1728.0
        
        self.Dmin_si = DSI_fromEng( self.Dmin )
        self.Dmax_si =  DSI_fromEng( self.Dmax )

        if do_print:
            minmaxL = ['Cond', "Cp", 'D', 'E', 'H', 'P', 'S', 'T', 'Visc']

            for name in minmaxL:
                max_name = name + 'max'
                print( '%10s'%max_name, '%9g'%getattr(self, max_name) )
                min_name = name + 'min'
                print( '%10s'%min_name, '%9g'%getattr(self, min_name) )
        # print( '.......................Finished calc_min_max_props ..........................')

        # Restore state to T and P from calling method
        self.setTP( T_save, P_save)        



def dev_tests():
    """
    __incompressibles_pure__ = ['Acetone', 'Air', 'AS10', 'AS20', 'AS30', 'AS40', 'AS55', 'DEB', 
    'DowJ', 'DowJ2', 'DowQ', 'DowQ2', 'DSF', 'Ethanol', 'ExampleDigitalPure', 'ExamplePure',
    'FoodAsh', 'FoodCarbohydrate', 'FoodFat', 'FoodFiber', 'FoodIce', 'FoodProtein', 'FoodWater', 
    'HC10', 'HC20', 'HC30', 'HC40', 'HC50', 'HCB', 'HCM', 'Hexane', 'HFE', 'HFE2', 'HY20', 'HY30',
    'HY40', 'HY45', 'HY50', 'LiqNa', 'NaK', 'NBS', 'PBB', 'PCL', 'PCR', 'PGLT', 'PHE', 'PHR', 
    'PLR', 'PMR', 'PMS1', 'PMS2', 'PNF', 'PNF2', 'S800', 'SAB', 'T66', 'T72', 'TCO', 'TD12', 
    'TVP1', 'TVP1869', 'TX22', 'TY10', 'TY15', 'TY20', 'TY24', 'Water', 'XLT', 'XLT2', 'ZS10', 
    'ZS25', 'ZS40', 'ZS45', 'ZS55']
    """

    symbol = 'Water'
    
    print( '='*22, "%s at Pressure = 0"%symbol, '='*22 )
    C = EC_Incomp_Fluid( symbol=symbol, T=None, P=0, auto_fix_value_errors=True )
    
    # C.setTP(T= (C.Tmin+C.Tmax)/2.0, P=100) # this temperature throws an exception ???
    
    C.printProps()

    if C.Tnbp is not None:
        print( '='*22, "%s at Tnbp"%symbol, 'Tnbp=%g degR'%C.Tnbp, '='*22 )
        C.setTP( C.Tnbp, 15.0)
        C.printProps()
    
        print( '='*22, "%s short prints at Tnbp"%symbol, '='*22 )
        print(C.getStrTransport())
        C.printTPD()



    print()
    
    print( '='*22, "%s over full Temp Range"%symbol, '='*22 )
    t_range = C.Tmax - C.Tmin
    N = 9
    tL = [C.Tmin + (i/N)*t_range for i in range(N+1)]
    print( ['%.1f'%T for T in tL])
    for T in tL:
        try:
            C.setTP(T, 5000.0)
            C.printTPD()
        except:
            print( 'T=%.1f FAILED'%T )
            Tsi = TSI_fromEng( T )
            Psi = PSI_fromEng( 50000 )

            Dsi = PropsSI('D','T',Tsi,'P',Psi,'INCOMP::%s'%symbol)
            print( '           Dsi =', Dsi)
        
    print()
    print( '='*22, "%s test constP_newH"%symbol, '='*22 )
    C.setTP(C.Tmid, 500.0)
    C.printTPD()
    C.constP_newH( C.H * 0.9 )
    C.printTPD()


    print()
    print( '='*22, "%s test constH_newP"%symbol, '='*22 )
    C.setTP(C.Tmid, 500.0)
    C.printTPD()
    C.constH_newP( 530.0 )
    C.printTPD()
    
    print()
    print( '='*22, "%s test setPH"%symbol, '='*22 )
    C.setTP(C.Tmid, 500.0)
    C.printTPD()
    C.setPH( 500., C.H * 0.95 )
    C.printTPD()
    
    print()
    print( '='*22, "%s test constS_newP"%symbol, '='*22 )
    C.setTP(C.Tmid, 500.0)
    C.printTPD()
    C.constS_newP(P=1000.0)
    C.printTPD()
    
    print()
    print( '='*22, "%s test setPD"%symbol, '='*22 )
    C.setTP(C.Tmid, 500.0)
    C.printTPD()
    C.setPD(P=800.0,D=C.D)
    C.printTPD()

    print()
    print( '='*22, "%s test calc_min_max_props"%symbol, '='*22 )
    C.calc_min_max_props( do_print=True )    
    print()
    print( '='*22, "print full %s properties"%symbol, '='*22 )
    C.printProps()
    print()

    banner( 'This should fail due to "auto_fix_value_errors = False"' )
    print( '='*22, "print full %s properties"%symbol, '='*22 )
    try:
        C_bad = EC_Incomp_Fluid( symbol='Water', T=None, P=0 )
    except:
        tb_str = traceback.format_exc()
        print( tb_str.split('raise ValueError')[-1])
        # print( tb_str )
    


if __name__ == '__main__':
    dev_tests()



from engcoolprop.ec_incomp_fluid import EC_Incomp_Fluid

# Create incompressible object. (without specifying state point)
ec_inc = EC_Incomp_Fluid(symbol="Acetone" )

# Print state point
ec_inc.printProps()



#!/usr/bin/env python
# -*- coding: ascii -*-
from __future__ import print_function

r"""
??? Tfreeze ???
EngCoolProp is a python Engineering Units wrapper around the CoolProp project.

<Paragraph description see docstrings at http://www.python.org/dev/peps/pep-0257/>
EngCoolProp uses units of primarily inch, lbm, lbf, sec, BTU (some use of ft and hour).

    The following are the default units for each property.

    T = Temperature = degR
    Tc= Critical Temperature = degR 
    Tnbp = Normal Boiling Point = degR
    P = Pressure = psia
    Pc = Critical Pressure = psia
    D = Density = lbm/cu ft
    rho = Density = lbm/cu inch
    Dc = Critical Density = lbm/cu ft
    E = Internal Energy = BTU/lbm
    H = Enthalpy = BTU/lbm
    S = Entropy = BTU/lbm degR
    Cv = Heat Capacity (const. V) = BTU/lbm degR
    Cp = Heat Capacity (const. P) = BTU/lbm degR
    g = Ratio of Specific Heats (Cp/Cv) = (-)
    A = Sonic Velocity = ft/sec
    V = Viscosity = 1.0E5 * lbm/ft-sec
    C = Thermal Conductivity = BTU/ft-hr-R
    MW = Molecular Weight = lbm/lbmmole
    Q = Quality (mass fraction gas) = (-)
    Z = Compressibility() = (-)



EngCoolProp
Copyright (C) 2018  Applied Python

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

-----------------------

"""
import os
here = os.path.abspath(os.path.dirname(__file__))


# for multi-file projects see LICENSE file for authorship info
# for single file projects, insert following information
__author__ = 'Charlie Taylor'
__copyright__ = 'Copyright (c) 2018 Charlie Taylor'
__license__ = 'GPL-3'
exec( open(os.path.join( here,'_version.py' )).read() )  # creates local __version__ variable
__email__ = "cet@appliedpython.com"
__status__ = "4 - Beta" # "3 - Alpha", "4 - Beta", "5 - Production/Stable"

#
import os
from CoolProp.CoolProp import PropsSI
from CoolProp import AbstractState
import CoolProp.CoolProp as CP

# Q values to indicate all liquid or gas
Q_LIQUID = 0  # all liquid
Q_GAS = 1     # all gas

def Deng_fromSI( D ): # 1 lbm/ft^3 = 16.01843417 kg/m^3
    return D  / 16.01843417

def DSI_fromEng( D ):
    return D * 16.01843417

def Teng_fromSI( T ):
    return T * 1.8

def TSI_fromEng( T ):
    return T / 1.8

def Peng_fromSI( P ):
    return P / 6894.76

def PSI_fromEng( P ):
    return P * 6894.76

def UHeng_fromSI( UH ): # 1 BTU/lbm = 2326 J/kg
    return UH / 2326.0

def UHSI_fromEng( UH ):
    return UH * 2326.0
    
def Seng_fromSI( S ):
    return S * 0.0002388461111111111

def SSI_fromEng( S ):
    return S / 0.0002388461111111111

def CPeng_fromSI( S ):
    return S * 0.0002388461111111111

def CPSI_fromEng( S ):
    return S / 0.0002388461111111111

def Aeng_fromSI( A ):
    return A * 3.28084
    
def ASI_fromEng( A ):
    return A / 3.28084

def Veng_fromSI( V ):
    # AS.viscosity() returns Pa*s
    return V * 0.6719689751 # convert from Pa*s to lbm/ft/sec
    
def VSI_fromEng( V ):
    return V / 0.6719689751

def CondEng_fromSI( Cond ):
    return Cond * 0.578176

def CondSI_fromEng( Cond ):
    return Cond / 0.578176

def EchoInput( x ):
    return x

# ========== build up generic call logic for "setProps" =============
# (i.e. discover which input pairs are supported in T,P,D,Q,S,H,E)

# Make so both:
# setProps(T=530, P=100)  AND  
# setProps(P=100, T=530) will work

# map property names between EngCoolProp and CoolProp
cool_varD = {} # index=eng, value=coolprop  (e.g.  cool_varD['H'] = 'Hmass')
cool_varD['H'] = 'Hmass'
cool_varD['D'] = 'Dmass'
cool_varD['S'] = 'Smass'
cool_varD['E'] = 'Umass'
cool_varD['T'] = 'T'
cool_varD['P'] = 'P'
cool_varD['Q'] = 'Q'

# map Eng to SI conversion functions for each of the fluid properties
toSI_callD = {} # index=eng, value=conversion func (e.g. SSI_fromEng)
toSI_callD['H'] = UHSI_fromEng
toSI_callD['D'] = DSI_fromEng
toSI_callD['S'] = SSI_fromEng
toSI_callD['E'] = UHSI_fromEng
toSI_callD['T'] = TSI_fromEng
toSI_callD['P'] = PSI_fromEng
toSI_callD['Q'] = EchoInput

# map SI to Eng conversion functions for each of the fluid properties
toEng_callD = {} # index=si char, value=conversion func (e.g. Seng_fromSI)
toEng_callD['H'] = UHeng_fromSI
toEng_callD['D'] = Deng_fromSI
toEng_callD['S'] = Seng_fromSI
toEng_callD['U'] = UHeng_fromSI
toEng_callD['T'] = Teng_fromSI
toEng_callD['P'] = Peng_fromSI
toEng_callD['Q'] = EchoInput
toEng_callD['C'] = CPeng_fromSI
toEng_callD['V'] = Veng_fromSI
toEng_callD['L'] = CondEng_fromSI


# create simple look-up that is order-independent for input pairs
call_tuplesD = {} # index=(ec1, ec2), value=(XX_INPUTS, c1, c2)

for s in dir(CP):
    if s.endswith('_INPUTS'):
        if s.find('molar')<0:
            name = s[:-7]
            for ec1, c1 in list(cool_varD.items()):
                if name.startswith(c1):
                    for ec2, c2 in list(cool_varD.items()):
                        if name.endswith(c2):
                            # put in both orderings of input pairs
                            call_tuplesD[(ec1, ec2)] = (getattr(CP, s), ec1, ec2)
                            call_tuplesD[(ec2, ec1)] = (getattr(CP, s), ec1, ec2)
            
# key = common names, value= CoolProp name
fluidNames = {"Acetone":"Acetone", 
              "Ammonia":"NH3", 
              "AR":"AR", 
              "Argon":"AR", 
              "Benzene":"Benzene", 
              "Butane":"Butane", 
              "Butene":"Butene", 
              "C(CH3)4":"Neopentane", 
              "C2H6O":"C2H6O", 
              "C3H6O":"Acetone", 
              "C6H6":"Benzene", 
              "Carbon Dioxide":"CO2", 
              "Carbon Monoxide":"CO", 
              "Carbonyl Sulfide":"COS",
              "CH(CH3)3":"Isobutane", 
              "CH2=C(CH3)2":"Isobutene", 
              "CH2=CH-CH3":"Propylene", 
              "CH2=CH2":"Ethylene", 
              "CH2CCH2":"Propyne", 
              "CH3-10(CH2)-CH3":"Dodecane", 
              "CH3-2(CH2)-CH3":"Butane", 
              "CH3-3(CH2)-CH3":"Pentane", 
              "CH3-4(CH2)-CH3":"Hexane", 
              "CH3-5(CH2)-CH3":"Heptane", 
              "CH3-6(CH2)-CH3":"Octane", 
              "CH3-7(CH2)-CH3":"Nonane", 
              "CH3-8(CH2)-CH3":"Decane", 
              "CH3-C6H5":"Toluene", 
              "CH3-CH2-CH=CH2":"Butene", 
              "CH3-CH=CH-CH3":"T2BUTENE", 
              "CH3CH2CH3":"Propane", 
              "CH3CH3":"Ethane", 
              "CH3OH":"Methanol", 
              "CH4":"CH4", 
              "CO":"CO", 
              "CO2":"CO2", 
              "COS":"COS", 
              "CYCLO-C3H6":"Cyclopropane", 
              "CYCLO-C6H12":"Cyclohexane", 
              "Cyclohexane":"Cyclohexane", 
              "Cyclopropane":"Cyclopropane", 
              "D2":"D2", 
              "D2O":"D2O", 
              "Decane":"Decane", 
              "Deuterium":"D2", 
              "Dodecane":"Dodecane", 
              "Ethane":"Ethane", 
              "ETHANOL":"ETHANOL", 
              "Ethanol":"ETHANOL", 
              "Ethylene":"Ethylene", 
              "F2":"Fluorine", 
              "Fluorine":"Fluorine", 
              "H2":"H2", 
              "H2O":"H2O", 
              "H2S":"H2S", 
              "HE":"HE", 
              "Heavy Water":"D2O", 
              "Helium":"HE", 
              "Heptane":"Heptane", 
              "Hexane":"Hexane", 
              "Hydrogen Sulfide":"H2S", 
              "Isobutane":"Isobutane", 
              "Isobutene":"Isobutene", 
              "Isohexane":"Isohexane", 
              "Isopentane":"Isopentane", 
              "KR":"Krypton", 
              "Krypton":"Krypton", 
              "Methane":"CH4", 
              "Methanol":"Methanol", 
              "N2":"N2", 
              "N2O":"N2O", 
              "NE":"Neon", 
              "Neon":"Neon", 
              "Neopentane":"Neopentane", 
              "NH3":"NH3", 
              "Nitrogen":"N2", 
              "Nitrous Oxide":"N2O", 
              "Nonane":"Nonane", 
              "O2":"O2", 
              "Octane":"Octane", 
              "Oxygen":"O2", 
              "ParaHydrogen":"ParaHydrogen", 
              "Pentane":"Pentane", 
              "PH2":"ParaHydrogen", 
              "Propane":"Propane", 
              "Propylene":"Propylene", 
              "Propyne":"Propyne", 
              "SF6":"SF6", 
              "SO2":"SO2", 
              "Sulfur Dioxide":"SO2", 
              "Sulfur Hexafluoride":"SF6", 
              "Toluene":"Toluene", 
              "Trans-Butene":"T2BUTENE", 
              "Water":"H2O", 
              "XE":"XE", 
              "Xenon":"XE", }

# make upper case key values for all entries
keyL = list(fluidNames.keys())     # can be legal or illegal name
valueL = list(fluidNames.values()) # all legal names
for v in valueL:
    fluidNames[v.upper()] = v # legal to legal lookup
for k in keyL:
    fluidNames[k.upper()] = fluidNames[k] # any key to legal lookup

class EC_Fluid(object):
    
    fluidNames = fluidNames

    def __init__(self,symbol="N2",T=530.0,P=1000.0, child=1):
        '''Init generic Fluid'''
        
        self.symbol = symbol.upper()
        
        # http://www.coolprop.org/_static/doxygen/html/class_cool_prop_1_1_abstract_state.html
        AS = AbstractState("HEOS", self.symbol)
        self.AS = AS
        
        self.name = AS.name()
        self.T = T
        self.P = P
        self.child = child
        
        Tsi = TSI_fromEng( T )
        Psi = PSI_fromEng( P )
        AS.update(CP.PT_INPUTS, Psi, Tsi)

        self.WtMol = AS.molar_mass() * 1000.0
        self.Tc = Teng_fromSI( AS.T_critical() )
        self.Pc = Peng_fromSI( AS.p_critical() )
        self.Dc = Deng_fromSI( AS.rhomass_critical() )
        self.Ttriple = Teng_fromSI( AS.Ttriple() )
        try:
            self.Tfreeze = Teng_fromSI( AS.T_freeze() )
        except:
            self.Tfreeze = self.Ttriple


        self.Tmin =  Teng_fromSI( AS.Tmin() )
        self.Tmax =  Teng_fromSI( AS.Tmax() )
        #self.Pmin = Peng_fromSI( AS.pmin() ) # missing from AbstractState
        self.Pmax = Peng_fromSI( AS.pmax() )
        
        dcIdeal = self.Pc*self.WtMol/self.Tc/10.729
        self.Zc = dcIdeal / self.Dc
        
        try:
            TnbpSI = PropsSI("T", "P", 101325, "Q", Q_LIQUID, self.symbol)
            self.good_nbp = True
            self.Tnbp = Teng_fromSI( TnbpSI )
        except:
            #print('WARNING... "%s" failed Normal Boiling Point Calculation.'%self.symbol)
            Ttriple = PropsSI(self.symbol,'Ttriple')
            #print('    Using Triple Point = %g degK as Tref'%Ttriple)
            self.good_nbp = False
            self.Tnbp = 'N/A'
            
        # set Tref value (used to get Href)
        if self.good_nbp:
            if self.Tnbp < 536.67:
                self.Tref = self.Tnbp # if NBP is low, use NBP as ref
            else:
                self.Tref = 536.67 # 536.67R = (SATP, Standard Ambient T P) = 77F, 25C 
        else:
            self.Tref = Teng_fromSI( Ttriple ) + 0.1
        self.Pref = 14.7
        
        #print( 'About to call setTP #1 with Tref, Pref=',self.Tref,self.Pref )
        self.setTP(self.Tref,self.Pref)
        #print( 'Back from call setTP')
        self.Href = self.H
        #print( 'About to call setTP #2')
        self.setTP(T,P)
        #print( 'Back from call setTP')
            
        if child==1: 
            self.dup = EC_Fluid(symbol=self.symbol,T=self.T,P=self.P, child=0)
            
        self.calcdFreezePt = 0

    def setProps(self, **inpD):
        '''Generic call using any two of supported inputs T,P,D,Q,S,H,E::
        
        #: for example
        #: ec.setProps(T=100, D=0.1)
        #: ec.setProps(D=0.1, T=100)
        #: ec.setProps(P=100, H=20)
        '''
        
        key = tuple( inpD.keys() )
        if key not in call_tuplesD:
            print('ERROR... setProps called with unknown inputs', inpD)
            print('Allowable inputs are: T,P,D,Q,S,H,E')
            if len(inpD) != 2:
                print('ONLY 2 inputs are allowed.  (e.g. T=530.0,P=100.0)')
            raise ValueError('setProps called with illegal inputs: %s'%str(inpD))
        
        
        INPUTS, ec1, ec2 = call_tuplesD[key]
        P1si = toSI_callD[ec1]( inpD[ec1] )
        P2si = toSI_callD[ec2]( inpD[ec2] )
        #print('Calling',key,'with',call_tuplesD[key],'P1si=%g,  P2si=%g'%(P1si, P2si))
        
        self.AS.update(INPUTS, P1si, P2si)
        
        self.setPropsFromAS()

    def dH_FromHref(self):
        """Calculate Enthalpy difference from self.Href"""
        return self.H - self.Href

    def setPropsFromAS(self):
        """Assume that self.AS (AbstractState) has been set elsewhere."""
        
        AS = self.AS

        self.T = Teng_fromSI( AS.T() )
        self.P = Peng_fromSI( AS.p() )

        self.H = UHeng_fromSI( AS.hmass() )
        self.D = Deng_fromSI( AS.rhomass() )
        self.rho = self.D / 1728.0
        #print( "rho=",rho,"in setTP")
        self.E = UHeng_fromSI( AS.umass() )
        self.S = Seng_fromSI( AS.smass() )
        self.Cv = CPeng_fromSI( AS.cvmass() )
        self.Cp = CPeng_fromSI( AS.cpmass() )
        self.sonicV = Aeng_fromSI( AS.speed_sound() )
        
        try:
            self.Visc = Veng_fromSI( AS.viscosity() ) * 1.0E5
        except:
            self.Visc = 0.0
        try:
            self.Cond = CondEng_fromSI( AS.conductivity() )
        except:
            self.Cond = 0.0
        self.Q = AS.Q()  # maybe use AS.PIP() ???
        
        dIdeal = self.P*self.WtMol/self.T/10.729
        self.Z = dIdeal / self.D

    def setTP(self,T=530.0,P=1000.0):
        '''Calc props from T and P'''
        
        Tsi = TSI_fromEng( T )
        Psi = PSI_fromEng( P )
        self.AS.update(CP.PT_INPUTS, Psi, Tsi)
        
        self.setPropsFromAS()
    
    def newDE(self,D=0.1,E=50.0):
        '''Calc properties at new D and E'''
        Dsi = DSI_fromEng( D )
        Esi = UHSI_fromEng( E )

        try:
            self.AS.update(CP.DmassUmass_INPUTS, Dsi, Esi)
            self.setPropsFromAS()
        except:
            #print('Failed in newDE for D=%g, E=%g'%(D, E))
            '''iterate on temperature until internal energy is correct
            tRbegin--beginning search temperature [R]'''
            
            if self.Ttriple > self.T:
                Tbegin = self.Ttriple
            else:
                Tbegin = self.T
            
            tolr = 1.0E-8
            tR = Tbegin
            #print '---'
            for i in range(48): # limit number of iterations
                self.setTD(T=tR, D=D)
                dedt = self.Cv
                eError = E-self.E
                
                #print 'dedt=',dedt,'  eError=',eError,'  tR=',tR
                
                if self.Cv==0.0:
                    print( '==> ERROR in wrap_dll.  CV=0 for tR=',tR )
                
                tR = tR + eError / dedt
                
                if abs(eError) <= tolr:
                    break
            
        
        
        
    def constP_newH(self,H):
        '''Calc properties at new H with same P'''
        
        Psi = PSI_fromEng( self.P )
        Hsi = UHSI_fromEng( H )
        self.AS.update(CP.HmassP_INPUTS, Hsi, Psi)
        
        self.setPropsFromAS()
        
    def constH_newP(self,P=1000.0):
        '''Calc properties at new P with same H'''
        
        Psi = PSI_fromEng( P )
        Hsi = UHSI_fromEng( self.H )
        
        self.AS.update(CP.HmassP_INPUTS, Hsi, Psi)
        self.setPropsFromAS()
        
    def setPH(self,P,H):
        '''Calc properties at P and H'''
        
        Psi = PSI_fromEng( P )
        Hsi = UHSI_fromEng( H )
        
        self.AS.update(CP.HmassP_INPUTS, Hsi, Psi)
        self.setPropsFromAS()

    def constS_newP(self,P=1000.0):
        '''Calc properties at new P with same S'''
        
        Psi = PSI_fromEng( P )
        Ssi = SSI_fromEng( self.S )
        self.AS.update(CP.PSmass_INPUTS, Psi, Ssi)
        self.setPropsFromAS()

    def setTD(self,T=530.0,D=0.01):
        '''Calc P from T and D'''
        
        Tsi = TSI_fromEng( T )
        Dsi = DSI_fromEng( D )
        self.AS.update(CP.DmassT_INPUTS, Dsi, Tsi)
        self.setPropsFromAS()
        
    def setPD(self,P=1000.0,D=0.01):
        '''Calc props from P and D'''
        
        Psi = PSI_fromEng( P )
        Dsi = DSI_fromEng( D )
        self.AS.update(CP.DmassP_INPUTS, Dsi, Psi)
        self.setPropsFromAS()

    def getSatP(self, T=None):
        """Get Psat for Liquid at input T or self.T"""
        if not(T is None) and (T < self.Tc):
            Tsi = TSI_fromEng( T )
            
            self.dup.AS.update(CP.QT_INPUTS, Q_LIQUID, Tsi)
            Psat = Peng_fromSI( self.dup.AS.p() )
        else:
            print('ERROR... getSatP called with T>Tc.   T=%s,   Tc=%g'%(T, self.Tc))
            Psat = self.P
            
        return Psat

    def getSatPandDens(self, T=None):
        """Get Psat and rho for both Liquid & Gas at input T or self.T"""
        if not(T is None) and (T < self.Tc):
            Tsi = TSI_fromEng( T )
            
            
            self.dup.AS.update(CP.QT_INPUTS, Q_GAS, Tsi)
            Dgas = self.D
            
            self.dup.AS.update(CP.QT_INPUTS, Q_LIQUID, Tsi)
            Psat = Peng_fromSI( self.dup.AS.p() )
            Dliq = Deng_fromSI( self.dup.AS.rhomass() )
            
            self.getSatP(T)
        else:
            print('ERROR... getSatP called with T>Tc.   T=%s,   Tc=%g'%(T, self.Tc))
            Psat = self.P
            Dgas = Dliq = self.D
            
            
        return Psat, Dliq, Dgas
        
    def getSatT(self, P=None):
        """Get Tsat for Liquid at input P or self.P"""
        if not(P is None) and (P < self.Pc):
            Psi = PSI_fromEng( P )
            self.dup.AS.update(CP.PQ_INPUTS, Psi, 1)
            Tsat = Teng_fromSI( self.dup.AS.T() )
        else:
            print('ERROR... getSatT called with P>Pc.   P=%s,   Pc=%g'%(P, self.Pc))
            Tsat = self.T
        
        return Tsat
            
    #for s in dir(CP):
    #    if s.endswith('_INPUTS'):
    #        print s,

    def getSatTandDens(self,P=None):
        """Get Tsat and Density for both Liquid & Gas at input P or self.P"""
        if not(P is None) and (P < self.Pc):
            Psi = PSI_fromEng( P )
            
            self.dup.AS.update(CP.PQ_INPUTS, Psi, Q_GAS)
            Dgas = Deng_fromSI( self.dup.AS.rhomass() )

            self.dup.AS.update(CP.PQ_INPUTS, Psi, Q_LIQUID)
            Tsat = Teng_fromSI( self.dup.AS.T() )
            Dliq = Deng_fromSI( self.dup.AS.rhomass() )
        else:
            print('ERROR... getSatT called with P>Pc.   P=%s,   Pc=%g'%(P, self.Pc))
            Tsat = self.T
            Dgas = Dliq = self.D
        
        return Tsat, Dliq, Dgas

    def dHvap(self, P=None):
        '''calculate heat of vaporization at input P or self.P (BTU/lbm)'''
        
        if P is None:
            P = self.P
        
        if (P < self.Pc):
            
            Psi = PSI_fromEng( P )
            self.dup.AS.update(CP.PQ_INPUTS, Psi, Q_LIQUID)
            Hliq = UHeng_fromSI( self.dup.AS.hmass() )
            
            self.dup.AS.update(CP.PQ_INPUTS, Psi, Q_GAS)
            Hgas = UHeng_fromSI( self.dup.AS.hmass() )

            return Hgas - Hliq
        else:
            print('ERROR in dHvap P=%g, Pc=%g'%(P, self.Pc))
            return 0.0
    
    def surfTen(self):
        '''calculate surface tension, lbf / inch'''
        
        st = self.AS.surface_tension() # units = N/m
        try:
            st_eng = st * 0.00571014709755764 # convert to lbf / inch
        except:
            st_eng = st
        #print('st = %s N/m = %s'%(st, st_eng))
        return st_eng # return lbf / inch
        

    def restoreFromDup(self):
        '''restore properties from duplicate n_fluid'''
        self.T = self.dup.T
        self.P = self.dup.P
        self.D = self.dup.D
        self.rho = self.D / 1728.0
        self.E = self.dup.E
        self.H = self.dup.H
        self.S = self.dup.S
        self.Cp = self.dup.Cp
        self.Cv = self.dup.Cv
        self.sonicV = self.dup.sonicV
        self.Visc = self.dup.Visc
        self.Cond = self.dup.Cond
        self.Q = self.dup.Q

    def saveToDup(self):
        '''save properties to duplicate n_fluid'''
        self.dup.T = self.T
        self.dup.P = self.P
        self.dup.D = self.D
        self.dup.E = self.E
        self.dup.H = self.H
        self.dup.S = self.S
        self.dup.Cp = self.Cp
        self.dup.Cv = self.Cv
        self.dup.sonicV = self.sonicV
        self.dup.Visc = self.Visc
        self.dup.Cond = self.Cond
        self.dup.Q = self.Q

    def initFromObj(self, obj):
        '''initialize properties from another n_fluid object'''
        if  self.symbol.upper() == obj.symbol.upper():
            #self.setTD(T=obj.T,D=obj.D)
            self.T = obj.T
            self.P = obj.P
            self.D = obj.D
            self.rho = self.D / 1728.0
            self.E = obj.E
            self.H = obj.H
            self.S = obj.S
            self.Cp = obj.Cp
            self.Cv = obj.Cv
            self.sonicV = obj.sonicV
            self.Visc = obj.Visc
            self.Cond = obj.Cond
            self.Q = obj.Q
        else:
            raise Exception('Wrong fluid for initializing')


    def gamma(self):
        '''calculate ratio of specific heats from Cp and Cv'''
        try:
            if self.Cp <= 0.0 or self.Cv <=0.0:
                g = 1.0 / 0.0
            g = self.Cp / self.Cv
            return g
        except:
            return 1.000001

    def getStrTransport(self):
        '''create a string from the Transport properties'''
        return  "%s Cp=%6g Cv=%6g gamma=%.4f Visc=%6g ThCond=%6g" %\
        (self.symbol,self.Cp, self.Cv, self.gamma(), self.Visc, self.Cond)

    def getStrTPD(self):
        '''create a string from the TPDEHS properties'''
        return  "%s T=%6.1f P=%6.1f D=%.4f E=%6.2f H=%6.2f S=%.3f Q=%.2f" %\
        (self.symbol,self.T, self.P, self.D, self.E, self.H, self.S, self.Q)

    def getStrTPDphase(self):
        '''create a string from the TPDEHS properties'''
        return  "%s T=%6.1f P=%6.1f D=%.4f E=%6.2f H=%6.2f S=%.3f %s" %\
        (self.symbol,self.T, self.P, self.D, self.E, self.H, self.S, self.Qdescription())

    def printTPD(self):
        '''print a string from the TPDEHS properties'''
        print(self.getStrTPD())
        
    def Qdescription(self):
        '''CoolProp Q values areDifferent from RefProp::
        
        #:c == RefProp ==
        #:c        q--vapor quality on a MOLAR basis [moles vapor/total moles]
        #:c           q < 0 indicates subcooled (compressed) liquid
        #:c           q = 0 indicates saturated liquid
        #:c           q = 1 indicates saturated vapor
        #:c           q > 1 indicates superheated vapor
        #:c           q = -998 subcooled liquid, but quality not defined (p > Pc)
        #:c           q = 999 indicates supercritical state (t > Tc) and (p > Pc)
        #:
        #:c == CoolProp ==
        #:c        q--vapor quality on a MASS basis [mass vapor/total mass]
        #:c           q  indicates subcooled (compressed) liquid
        #:c           q  indicates saturated liquid
        #:c           q  indicates saturated vapor
        #:c           q < 0 indicates superheated vapor OR (p > Pc)
        #:c           q  subcooled liquid, but quality not defined (p > Pc)
        #:c           q  indicates supercritical state (t > Tc) and (p > Pc)
        #:
        #:Phase String            Phase Region
        #:liquid                  p < pcrit & T < Tcrit ; above saturation
        #:gas                     p < pcrit & T < Tcrit ; below saturation
        #:twophase                p < pcrit & T < Tcrit ; mixed liquid/gas
        #:
        #:supercritical_liquid    p > pcrit & T < Tcrit
        #:supercritical_gas       p < pcrit & T > Tcrit
        #:supercritical           p > pcrit & T > Tcrit        
        
        '''
        if self.P >= self.Pc: # supercritical pressure
            if self.T < self.Tc:
                return 'Supercritical Liquid'
            else:
                return 'Supercritical Fluid'
        else:
            # P < Pc
            if self.T >= self.Tc:
                return 'Supercritical Gas'
            else:
                # both T<Tc  AND  P<Pc
                if self.Q <= 0.0:
                    return "Liquid"
                elif self.Q > 900.0:
                    return "Supercrit"
                elif self.Q >= 1.0:
                    return "Gas"
                else:
                    pcent = self.Q * 100.0
                    return "%.2f "%pcent + " %gas"

    def html_desc(self):
        '''html output a multiline property summary with units'''
        return '<table><th colspan="3" align="left">&nbsp;&nbsp;&nbsp;for fluid "'+ str(self.name)+ " (" + str(self.symbol) + ')"</th>' \
            + "<tr><td>T = </td><td>" + "%5g"%self.T + "</td><td> degR &nbsp;&nbsp;(Tc=" + "%5g"%self.Tc + ")"+ "</td></tr>" \
            + "<tr><td>P = </td><td>" + "%5g"%self.P + "</td><td> psia &nbsp;&nbsp;&nbsp;(Pc=" + "%5g"%self.Pc + ")"+ "</td></tr>" \
            + "<tr><td>D = </td><td>" + "%5g"%self.D + "</td><td> lbm/cu ft &nbsp;&nbsp;(Dc=" + "%5g"%self.Dc + ")"+ "</td></tr>" \
            + "<tr><td>E = </td><td>" + "%5g"%self.E + "</td><td> BTU/lbm"+ "</td></tr>" \
            + "<tr><td>H = </td><td>" + "%5g"%self.H + "</td><td> BTU/lbm"+ "</td></tr>" \
            + "<tr><td>S = </td><td>" + "%5g"%self.S + "</td><td> BTU/lbm degR"+ "</td></tr>" \
            + "<tr><td>Cv= </td><td>" + "%5g"%self.Cv + "</td><td> BTU/lbm degR"+ "</td></tr>" \
            + "<tr><td>Cp= </td><td>" + "%5g"%self.Cp + "</td><td>BTU/lbm degR"+ "</td></tr>" \
            + "<tr><td>A = </td><td>" + "%5g"%self.sonicV + "</td><td> ft/sec"+ "</td></tr>" \
            + "<tr><td>MW= </td><td>" + "%5g"%self.WtMol + "</td><td> lbm/lbmmole"+ "</td></tr>" \
            + "<tr><td>Q = </td><td>" + "%s"%self.Qdescription() + "</td><td> phase"+ "</td></tr>" \
            + "<tr><td>Z = </td><td>" + "%5g"%self.Z + "</td><td>(-)"+ "</td></tr></table>"

    def printCriticalProps(self):
        '''print a multiline property summary with units'''
        print("Critical Properties for fluid",self.name,"("+self.symbol+")")
        print("Tc=%8g"%self.Tc,"degR")
        if self.good_nbp:
            print("Tnbp=%8g"%self.Tnbp,"degR")
        else:
            print("Tnbp=N/A")
            
        print("Pc =%8g"%self.Pc,"psia")
        print("Dc =%8g"%self.Dc,"lbm/cu ft")
        print("MW=%8g"%self.WtMol," lbm/lbmmole")
        print("Zc =%8g"%self.Zc," (-)")
        
        try:
            
            Tsi = TSI_fromEng( self.Tc )
            Dsi = DSI_fromEng( self.Dc )
            self.dup.AS.update(CP.DmassT_INPUTS, Dsi, Tsi)
            
            H = UHeng_fromSI( self.dup.AS.hmass() )
            print("Hc =%8g"%H," BTU/lbm", sep=' ')
            E = UHeng_fromSI( self.dup.AS.umass() )
            print("Ec =%8g"%E," BTU/lbm", sep=' ')
            S = Seng_fromSI( self.dup.AS.smass() )
            print("Sc =%8g"%S," BTU/lbm degR", sep=' ')
            #Cv = CPeng_fromSI( self.dup.AS.cvmass() )
            #print("Cvc=%8g"%Cv," BTU/lbm degR", sep=' ')
            #Cp = CPeng_fromSI( self.dup.AS.cpmass() )
            #print("Cpc=%8g"%Cp," BTU/lbm degR", sep=' ')
            sonicV = Aeng_fromSI( self.dup.AS.speed_sound() )
            print("Ac =%8g"%sonicV," ft/sec", sep=' ')
            Visc = Veng_fromSI( self.dup.AS.viscosity() ) * 1.0E5
            print("Vc =%8g"%Visc," viscosity [1.0E5 * lbm/ft-sec]", sep=' ')
            Cond = CondEng_fromSI( self.dup.AS.conductivity() )
            print("Cc =%8g"%Cond," thermal conductivity [BTU/ft-hr-R]", sep=' ')
        except:
            pass

    def printProps(self):
        '''print a multiline property summary with units'''
        print("State Point for fluid",self.name,"("+self.symbol+")", sep=' ')
        if self.good_nbp:
            print("T =%8g"%self.T," degR (Tc=%8g"%self.Tc,", Tnbp=%8g"%self.Tnbp, "Ttriple=%8g"%self.Ttriple,")", sep=' ')
        else:
            print("T =%8g"%self.T," degR (Tc=%8g"%self.Tc,", Tnbp=N/A", "Ttriple=%8g"%self.Ttriple,")", sep=' ')
            
        print("P =%8g"%self.P," psia (Pc=%8g"%self.Pc,")", sep=' ')
        print("D =%8g"%self.D," lbm/cu ft (Dc=%8g"%self.Dc,")", sep=' ')
        print("E =%8g"%self.E," BTU/lbm", sep=' ')
        print("H =%8g"%self.H," BTU/lbm", sep=' ')
        print("S =%8g"%self.S," BTU/lbm degR", sep=' ')
        print("Cv=%8g"%self.Cv," BTU/lbm degR", sep=' ')
        print("Cp=%8g"%self.Cp," BTU/lbm degR", sep=' ')
        print("g =%8g"%self.gamma()," Cp/Cv (-)", sep=' ')
        print("A =%8g"%self.sonicV," ft/sec", sep=' ')
        print("V =%8g"%self.Visc," viscosity [1.0E5 * lbm/ft-sec]", sep=' ')
        print("C =%8g"%self.Cond," thermal conductivity [BTU/ft-hr-R]", sep=' ')
        print("MW=%8g"%self.WtMol," lbm/lbmmole", sep=' ')
        print("Q =%8g"%self.Q," Vapor Quality (mass fraction gas)", sep=' ')
        print("Z =%8g"%self.Z," (-)", sep=' ')

    def compressibility(self):
        '''returns Z'''
        return self.Z

    def dH_FromHZero(self):
        return self.H - self.Href

if __name__ == '__main__':
    
    C = EC_Fluid( symbol='CO2' )
    
    # C.setProps(T=C.Tref, Q=0.5)
    #C.setProps(P=5., Q=0.5)
    #C.setProps(H=120., E=0.5) # illegal inputs
    
    C.printProps()
    print('='*55)
    
    C = EC_Fluid( symbol='N2' )
    
    
    print(C.getStrTransport())
    print(C.getStrTPDphase())
    C.printTPD()
    print('.'*55)
    C.printCriticalProps()
    print()
    #print( C.html_desc() )
    
    if 1:
        print('='*55)
        for T in [115., 150., 200., 250., 300., 350., 400., 500., 550.]:
            C.setTP(T, 100.0)
            C.printTPD()
            
    print('$'*55)
    C.newDE(D=1,E=50.0)
    C.printTPD()
    
    C.constP_newH( 80.0 )
    C.printTPD()
    C.constH_newP( 500.0 )
    C.printTPD()
    
    C.setPH( 300., 75.0 )
    C.printTPD()
    
    C.constS_newP(P=1000.0)
    C.printTPD()
    
    C.setTD(T=500.0,D=5.0)
    C.printTPD()
    
    C.setPD(P=800.0,D=4.0)
    C.printTPD()
    
    print('------- Psat Test ----------')
    Psat = C.getSatP( T=200.0 )
    print( 'Psat = %g psia'%Psat )
    C.setProps(Q=Q_LIQUID, P=Psat)
    C.printTPD()
    C.setProps(Q=Q_GAS, P=Psat)
    C.printTPD()
    print()
    print('Psat=%g, rhoLiq=%g, rhoGas==%g'%C.getSatPandDens(T=190.0))
    C.printTPD()
    
    Tsat = C.getSatT( P=150.0 )
    print('Tsat = %g degR'%Tsat)
    C.setProps(Q=Q_LIQUID, T=Tsat)
    C.printTPD()
    C.setProps(Q=Q_GAS, T=Tsat)
    C.printTPD()
    
    print()
    print('Tsat=%g, rhoLiq=%g, rhoGas==%g'%C.getSatTandDens( P=200.0 ))
    C.printTPD()
    C.getSatT( P=196.1 )
    C.printTPD()
    
    print('_'*55)
    Tsat = C.getSatT( P=14.7 )
    print('Tnpb =',C.Tnbp,'   Tsat @1atm =',Tsat)
    C.setTP(C.Tnbp, 14.7)
    for P in [5.0, 14.7, 50.0]:
        dH = C.dHvap( P=P )
        print('dHvap=%g at P=%g'%(dH,P) )
    
    
    



from engcoolprop.ec_incomp_fluid import EC_Incomp_Fluid

# Create incompressible object. (without specifying state point)
ec_inc = EC_Incomp_Fluid(symbol="Water")

# Print state point
ec_inc.printProps() # Print state point at given T,P

print()
ec_inc.printTPD()
print()
ec_inc.printTransport()

from engcoolprop.ec_incomp_fluid import EC_Incomp_Fluid

# Create incompressible object at T=500 degR, P=500 psia and max pressure = 5000 psia
ec_inc = EC_Incomp_Fluid(symbol="Water", T=500, P=500, Pmax=5000)

# Print state point
ec_inc.printProps()

ec_inc.setTP( 600, 600 )
ec_inc.printProps()



"""
In the CoolProp module, properties available for incompressible fluids include:

Temperature (T)
Pressure (P)
Density (D)
Specific Heat Capacity (C)
Internal Energy (U)
Enthalpy (H)
Entropy (S)
Viscosity (V)
Thermal Conductivity (L)
Minimum Temperature (Tmin)
Maximum Temperature (Tmax)

These properties can be accessed using the PropsSI function. 
CoolProp supports a variety of incompressible fluids, including pure fluids and binary mixtures.

Working independent pairs: TP, HP, SP, DP
from chk_indep_props_incomp.py:
    ONLY WORKS FOR: DP, HP, PS, PT  <-- NEED Pressure <--

NOT working: DT, HD, HS, UP, UD

"""
import CoolProp.CoolProp as CP

# Example fluid: Water
fluid = 'INCOMP::Water'

# Specifying Temperature and Pressure
T = 298.15  # Temperature in K
P = 101325  # Pressure in Pa

density = CP.PropsSI('D', 'T', T, 'P', P, fluid)
specific_heat_capacity = CP.PropsSI('C', 'T', T, 'P', P, fluid)
internal_energy = CP.PropsSI('U', 'T', T, 'P', P, fluid)
enthalpy = CP.PropsSI('H', 'T', T, 'P', P, fluid)
entropy = CP.PropsSI('S', 'T', T, 'P', P, fluid)
viscosity = CP.PropsSI('V', 'T', T, 'P', P, fluid)
thermal_conductivity = CP.PropsSI('L', 'T', T, 'P', P, fluid)

print(f"Density: {density} kg/m³")
print(f"Specific Heat Capacity: {specific_heat_capacity} J/kg.K")
print(f"Internal Energy: {internal_energy} J/kg")
print(f"Enthalpy: {enthalpy} J/kg")
print(f"Entropy: {entropy} J/kg.K")
print(f"Viscosity: {viscosity} Pa.s")
print(f"Thermal Conductivity: {thermal_conductivity} W/m.K")

# Specifying Density and Temperature
D = density  # Density in kg/m³

# DT NOT supported
# pressure_from_DT = CP.PropsSI('P', 'D', D, 'T', T, fluid)
# print(f"Pressure from Density and Temperature: {pressure_from_DT} Pa")

# Specifying Enthalpy and Pressure
H = enthalpy  # Enthalpy in J/kg

temperature_from_HP = CP.PropsSI('T', 'H', H, 'P', P, fluid)
print(f"Temperature from Enthalpy and Pressure: {temperature_from_HP} K")

# Specifying Entropy and Pressure
S = entropy  # Entropy in J/kg.K

temperature_from_SP = CP.PropsSI('T', 'S', S, 'P', P, fluid)
print(f"Temperature from Entropy and Pressure: {temperature_from_SP} K")

# Minimum and Maximum Temperature
Tmin = CP.PropsSI('Tmin', fluid)
Tmax = CP.PropsSI('Tmax', fluid)

print(f"Minimum Temperature: {Tmin} K")
print(f"Maximum Temperature: {Tmax} K")

# =============================================================


# Specifying Density and Pressure
D = 997  # Density in kg/m³
P = 101325  # Pressure in Pa

temperature_from_DP = CP.PropsSI('T', 'D', D, 'P', P, fluid)
print(f"Temperature from Density and Pressure: {temperature_from_DP} K")

# Specifying Enthalpy and Density
H = 419000  # Enthalpy in J/kg

# HD NOT supported
# temperature_from_HD = CP.PropsSI('T', 'H', H, 'D', D, fluid)
# print(f"Temperature from Enthalpy and Density: {temperature_from_HD} K")

# Specifying Enthalpy and Entropy
S = 1500  # Entropy in J/kg.K

# HS NOT supported
# temperature_from_HS = CP.PropsSI('T', 'H', H, 'S', S, fluid)
# print(f"Temperature from Enthalpy and Entropy: {temperature_from_HS} K")

# Specifying Internal Energy and Pressure
U = 210000  # Internal Energy in J/kg

# UP NOT supported
# temperature_from_UP = CP.PropsSI('T', 'U', U, 'P', P, fluid)
# print(f"Temperature from Internal Energy and Pressure: {temperature_from_UP} K")

# UD NOT supported
# Specifying Internal Energy and Density
# temperature_from_UD = CP.PropsSI('T', 'U', U, 'D', D, fluid)
# print(f"Temperature from Internal Energy and Density: {temperature_from_UD} K")




import traceback
from CoolProp.CoolProp import PropsSI
import CoolProp.CoolProp as CP
from engcoolprop.ec_fluid import ( Peng_fromSI, PropsSI , TSI_fromEng, PSI_fromEng, UHeng_fromSI,
                                   Teng_fromSI, Deng_fromSI, UHSI_fromEng, DSI_fromEng  )
# from engcoolprop.find_exception_threshold import find_exception_limit

from engcoolprop.ec_fluid import toSI_callD

from engcoolprop.root_solver import brentq

# ==================================================================================
def calc_Psat_psia( TdegR, fluid ):
    Psat_psia = Peng_fromSI( PropsSI('P','T', TSI_fromEng( TdegR ) ,'Q',0, fluid) )
    return Psat_psia

# ==================================================================================
def find_T_from_Dterp(ec_inc, Dtarg, tol=1e-12, max_iter=1000):
    """
    Using the Interpolator ec_inc.Dterp (D=f(T)), calculate T for a target D
    Dtarg = Lbm/cuft
    Return T = degR
    """


    D_low = ec_inc.Dmin
    D_high = ec_inc.Dmax

    if Dtarg < D_low:
        print( 'WARNING: Dtarg=%g too low in find_T_from_D. Returning Tmax=%g R'%( Dtarg, ec_inc.Tmax ))
        return ec_inc.Tmax, 0 # all good (returning TdegR)
    if Dtarg > D_high:
        print( 'WARNING: Dtarg=%g too high in find_T_from_D. Returning Tmin=%g R'%( Dtarg, ec_inc.Tmin ))
        return ec_inc.Tmin, 0 # all good (returning TdegR)
    # ............................................................

    def get_d_param_at_T( T ):
            D = ec_inc.Dterp( T )
            return D - Dtarg # finds 0 point

    sol = brentq(get_d_param_at_T, ec_inc.Tmin, ec_inc.Tmax, xtol=tol )

    TdegR = sol.root

    if sol.converged:
        return TdegR, 0 # all good
    else:
        return TdegR, 1 # error


# ==================================================================================
def find_T_from_D(ec_inc, D, tol=1e-7, max_iter=1000):
    # Define the search range for temperature in Kelvin
    # T_low_si = ec_inc.Tmin_si  # Lower bound of temperature in K
    # T_high_si = ec_inc.Tmax_si  # Upper bound of temperature in K

    Dtarg_si = DSI_fromEng( D  ) # Do calcs in SI

    D_low_si = ec_inc.Dmin_si
    D_high_si = ec_inc.Dmax_si

    if Dtarg_si < D_low_si:
        print( 'WARNING: Dtarg_si=%g too low in find_T_from_D. Returning Tmax=%g R'%( Dtarg_si, ec_inc.Tmax ))
        return ec_inc.Tmax, 0 # all good (returning TdegR)
    if Dtarg_si > D_high_si:
        print( 'WARNING: Dtarg_si=%g too high in find_T_from_D. Returning Tmin=%g R'%( Dtarg_si, ec_inc.Tmin ))
        return ec_inc.Tmin, 0 # all good (returning TdegR)
    # ............................................................

    def get_d_param_at_T( Tsi ):
        if 1:#try:
            Dsi = CP.PropsSI('D', 'T', Tsi, 'P', ec_inc.Pmax_si, ec_inc.fluid)
            return Dsi - Dtarg_si # finds 0 point
        # except:
        #     return float('inf')

    sol = brentq(get_d_param_at_T, ec_inc.Tmin_si, ec_inc.Tmax_si, xtol=tol )

    TdegR = Teng_fromSI(sol.root)

    if sol.converged:
        return TdegR, 0 # all good
    else:
        return TdegR, 1 # error




# ==================================================================================
def find_T_at_P( ec_inc, P, dep_name='H', dep_val=0, tol=1.0E-6):
    """
    Iterate on T to find the value of the dependent variable for given P
    """

    TargSI = toSI_callD[dep_name]( dep_val ) # get target from Eng units to SI

    def get_opt_targ_of_T( T ): # T in degR
        Psat = ec_inc.get_Psat( T )
        Psi = PSI_fromEng( max(P, Psat + 0.0000001) ) # force fluid to Liquid if below Psat line
        try:
            SI_prop = PropsSI(dep_name, 'T',TSI_fromEng(T),'P',Psi,'INCOMP::%s'%ec_inc.symbol)
            return SI_prop - TargSI # finds 0 point
        except:
            return float('inf')

    sol = brentq(get_opt_targ_of_T, ec_inc.Tmin, ec_inc.Tmax, xtol=tol )

    # print( "sol.function_calls=%s, sol.iterations=%s"%(sol.function_calls, sol.iterations),
    #        'for', "dep_name=%s, dep_val=%s"%(dep_name, dep_val) )

    TdegR = sol.root

    if sol.converged:
        return TdegR, 0 # all good
    else:
        return TdegR, 1 # error


# ==================================================================================
def calc_T_freeze( ec_inc  ):
    """ 
    Use brentq method to solve for T_freeze
    (code taken from scipy)
    see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root_scalar.html
    or: https://en.wikipedia.org/wiki/Brent%27s_method
    for background info.
    """

    # find Tnbp where Psat = 1 atm (14.6959 psia)

    Tmin_si =   PropsSI('Tmin','T',0,'P',0,'INCOMP::%s'%ec_inc.symbol) * 1.0000000000000002
    Tmax_si =   PropsSI('Tmax','T',0,'P',0,'INCOMP::%s'%ec_inc.symbol) * 0.9999999999999999
    Tmin = Teng_fromSI( Tmin_si )
    Tmax = Teng_fromSI( Tmax_si )

    T = int( Tmin + 1)
    Psi = PSI_fromEng( 5000 )

    target_str = 'below the freezing point of'

    while T < Tmax:
        try:
            PropsSI('D', 'T',TSI_fromEng(T),'P',Psi, ec_inc.fluid)
        except:
            tb_str = traceback.format_exc()
            if target_str in tb_str:
                sL = tb_str.split( target_str )
                sL = sL[1].split()
                s = sL[0][:-1]
                try:
                    T_freeze = float( s.strip() )
                    return T_freeze
                except:
                    pass
        T += 1

    return 0



# ==================================================================================
def calc_Tnbp( ec_inc, method=None  ):
    """ 
    Use brentq method to solve for Tnbp
    (code taken from scipy)
    see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root_scalar.html
    or: https://en.wikipedia.org/wiki/Brent%27s_method
    for background info.
    """
    # only calc Tnbp if Psat is supported for ec_inc
    if not  ec_inc.get_Psat( ec_inc.Tmax ) > 0:
        return 0.0, 1 # error

    # find Tnbp where Psat = 1 atm (14.6959 psia)

    f = lambda T: calc_Psat_psia( T, ec_inc.fluid) - 14.6959 # find t(T) == 0

    tol=1.0e-6
    sol = brentq(f, ec_inc.Tmin, ec_inc.Tmax, xtol=tol )

    Tnbp = sol.root

    # psia = calc_Psat_psia( Tnbp, ec_inc.fluid)
    # print( "Tnbp=%.1f, psia=%.1f, sol.method=%s"%(Tnbp, psia, sol.method) )

    if sol.converged:
        return Tnbp, 0 # all good
    else:
        return Tnbp, 1 # error

# ==================================================================================
if __name__ == "__main__":

    from engcoolprop.ec_incomp_fluid import EC_Incomp_Fluid

    ec_inc = EC_Incomp_Fluid(symbol="DowQ", P=100.0, show_warnings=1 ) # T=degR, P=psia

    Tnbp, error_code = calc_Tnbp( ec_inc, method=None )

    if not error_code:
        print( 'For fluid = %s, Tnbp = %g degR'%( ec_inc.fluid, Tnbp ) )
    else:
        print( 'For fluid = %s, Tnbp calculation FAILED.'%ec_inc.fluid )
    print( '='*22, "Tnbp Calcs", '='*22 )

    EPSILON = 1.0e-5

    N = 10
    t_range = ec_inc.Tmax - ec_inc.Tmin
    tL = [ec_inc.Tmin + (i/N)*t_range for i in range(N+1)] 

    p_range = ec_inc.Pmax - ec_inc.Pmin
    pL = [ec_inc.Pmin + (i/N)*p_range for i in range(N+1)] 

    print( '='*22, "T from H Calcs", '='*22 )
    for T in tL:
        for P in pL:
            
            ec_inc.setTP( T, P)

            Htarg = ec_inc.H 
            TdegR, err_flag =find_T_at_P( ec_inc, P, dep_name='H', dep_val=Htarg)

            error = T - TdegR
            if abs(error) > EPSILON:
                print( 'T=%g'%TdegR, 'for Htarg=%g'%Htarg, ' and P=%g'%P, "T=%g"%T , 'error=', error, 'err_flag=', err_flag)


    print( '='*22, "T from D Calcs", '='*22 )
    for T in tL:
        for P in pL:
            
            ec_inc.setTP( T, P)

            TdegR, err_flag = find_T_from_Dterp(ec_inc, ec_inc.D, max_iter=1000)

            error = T - TdegR
            if abs(error) > EPSILON:
                print( 'T=%g'%TdegR, 'for Dtarg=%g lbm/cuft'%ec_inc.D, ' and P=%g'%P, "T=%g"%T , 'error=', error, 'err_flag=', err_flag)

    # Dtarg = ec_inc.D
    # P = 70
    # TdegR, err_flag =find_T_at_P( ec_inc, P, dep_name='D', dep_val=Dtarg)
    # print( 'T=%g'%TdegR, 'for Dtarg=%g'%Dtarg, ' and P=%g'%P)

    # ec_inc.setTP( TdegR, P)
    # ec_inc.printProps()







import CoolProp.CoolProp as CP
import traceback

N2 = CP.AbstractState("HEOS", 'N2')

DowQ = CP.AbstractState("INCOMP", 'DowQ')
DowQ.update(CP.PT_INPUTS, 101325, 500)

n2L = dir( N2 )
dowqL = dir( DowQ )

# print( 'n2L =', n2L )
print()
# print( 'dowqL =', dowqL )

for d_str in dowqL:
    d = getattr( DowQ, d_str )
    if not d_str.startswith('__'):
        try:
            print( d_str, 'd() =', d() )
            print( '...........................................')
        except:
            if 0:
                tb_str = traceback.format_exc()
                if tb_str.find('not implement') < 0 :
                    print( d_str, type(d))
                    print( 'FAILED', d_str)
                    print( tb_str )
                    print( '...........................................')

# print( '... .molar_mass() ...')
# print( N2.molar_mass() )
# print( DowQ.molar_mass() )






import traceback
from CoolProp.CoolProp import PropsSI
from engcoolprop.find_exception_threshold import find_exception_limit
# from engcoolprop.parameter_units import si_unitsD
from engcoolprop.ec_fluid import  Peng_fromSI, PSI_fromEng

prop_descD = {} # key:eng prop desc, value:coolprop prop desc
prop_descD ['T'] = 'T'
prop_descD ['P'] = 'P'
prop_descD ['D'] = 'D'
prop_descD ['E'] = 'U'
prop_descD ['H'] = 'H'
prop_descD ['S'] = 'S'
prop_descD ['Cp'] = 'C'
prop_descD ['Visc'] = 'V'
prop_descD ['Cond'] = 'L'

# NOTE: 'P' is always an independent variable
LEGAL_PROP_CHR = set( ['T', 'D', 'U', 'H', 'S', 'C', 'V', 'L', 'P'] )

def is_frac_max_check_inclusive( base_name ):
    """ 
    Some Solutions have problem with non-inclusive check of range.
    Check to if this "base_name" solution has that problem. 
    """

    # get frac_max from CoolProp DB
    frac_max = PropsSI('fraction_max','INCOMP::%s'%base_name) 
    
    # create max fraction input from base name of solution
    symbol = base_name + '-' + '%g'%( frac_max*100 ) + '%'

    # Use Tmax as temperature for density calc
    T = PropsSI('Tmax','T',0,'P',0, 'INCOMP::%s'%symbol)
    P = 34.473E+6 # Pa (5000 psia)

    try:
        PropsSI('D', 'T',T,'P',P, 'INCOMP::%s'%symbol)
        all_good = True
    except:
        all_good = False
    return all_good


def get_si_prop( Psi_val, targ_prop='H', ind_name='T', ind_si_val=1000.0, symbol='Water'):
    """
    Psi_val (float): Pressure in SI units (Pa)
    target_prop (string): property to be calculated (e.g. 'T', 'D', 'U', 'H', 'S', 'C', 'V', 'L')
    ind_name (string): independent variable name (e.g. 'T', 'D', 'H', 'S')
    ind_si_val (float): independent variable value in SI units (e.g. K, kg/m³, J/kg, J/kg.K)
    symbol (string): symbol of incompressible fluid

    NOTE: this will often throw an Exception if the pressure is below saturation pressure
    """
    # print( 'Psi_val, targ_prop, ind_name, ind_si_val =', Psi_val, targ_prop, ind_name, ind_si_val )
    
    SI_prop = PropsSI(targ_prop, ind_name,ind_si_val,'P',Psi_val,'INCOMP::%s'%symbol)
    return SI_prop


def safe_get_INCOMP_prop( prop_desc, Psi_val=100000, ind_name='T', ind_si_val=1000.0, 
                          symbol='Water', show_warnings=True, Pmax=10000 ):
    """
    Return desired property from PropsSI for INCOMP fluid with Exception protection
    from an invalid Pressure. (e.g. below saturation pressure)

    Args:
        prop_desc (string): property to calculate: can be T, D, E, H, S, Cp, Visc, Cond    
        Psi_val (float): SI pressure value in "Pa"
        ind_name (string): independent variable other than P. (i.e. D, H, S, T)
        ind_si_val (float): independent variable value 
            D = "kg/m^3"
            H = "J/kg"
            S = "J/kg/K"
            T = "K"
        show_warnings (bool): will show warnings if == True
        Pmax (float): psia max pressure searched 
            
    ...CoolProp Incompressible Properties...
    Temperature (T): Kelvin (K)
    Pressure (P): Pascal (Pa)
    Density (D): Kilograms per cubic meter (kg/m³)
    Specific Heat Capacity (C): Joules per kilogram per Kelvin (J/kg.K)
    Internal Energy (U): Joules per kilogram (J/kg)
    Enthalpy (H): Joules per kilogram (J/kg)
    Entropy (S): Joules per kilogram per Kelvin (J/kg.K)
    Viscosity (V): Pascal-seconds (Pa.s)
    Thermal Conductivity (L): Watts per meter per Kelvin (W/m.K)
    """

    # if Psi_val == 0.0:
    #     print( 'Entered safe_get... at Psi_val=',Psi_val, "Peng_fromSI(Psi_val)=", Peng_fromSI(Psi_val))

    # get legal CoolProp description from a more generous prop_descD dictionary
    if prop_desc in prop_descD:
        prop_chr = prop_descD[ prop_desc ] # e.g. prop_descD ['Cond'] = 'L'
    else:
        prop_chr = prop_desc

    # demand that prop_chr is in legal CoolProp values
    if prop_chr not in  LEGAL_PROP_CHR: # i.e. in set( ['T', 'D', 'U', 'H', 'S', 'C', 'V', 'L', 'P'] )
        raise ValueError( 'prop_chr "%s" must be in '%prop_chr + repr(LEGAL_PROP_CHR) )
    
    
    try:
        prop_val =  get_si_prop( Psi_val, targ_prop=prop_desc, ind_name=ind_name, ind_si_val=ind_si_val, symbol=symbol )
        good_Psi_val = Psi_val
    except:
        # If there is an exception, assume it is a pressure issue.
        # Find the pressure that just works (i.e. right on the Exception threshold)
        # print( 'Exception at Psi_val=',Psi_val, "Peng_fromSI(Psi_val)=", Peng_fromSI(Psi_val))
        # tb_str = traceback.format_exc()
        # print( tb_str )
        
        f = lambda P: get_si_prop( P, targ_prop=prop_desc, ind_name=ind_name, ind_si_val=ind_si_val, symbol=symbol )

        good_Psi_val = find_exception_limit(f, tolerance=1e-4, lower_bound=1.0e-9, upper_bound=PSI_fromEng(Pmax),
                                             show_warnings=show_warnings )

        # start interpreting the output of find_exception_limit
        psia_in = Peng_fromSI(Psi_val)
        if good_Psi_val is None:
            if show_warnings:
                print( 'REALLY bad error: good_Psi_val is None')
                # print( traceback.format_exc() )
            psia_2 = 10000.0 # psia
        else:
            psia_2  = Peng_fromSI(good_Psi_val)

        if show_warnings and Peng_fromSI(good_Psi_val - Psi_val) > 0.1: # only show warning if error is significant
            s = ind_name + 'P'
            print( s,'WARNING: P override (%s'%Psi_val + '-->%s Pa)'%good_Psi_val, 
                '(%.1f-->%.1f psia)'%(psia_in, psia_2), 'calculating:',prop_desc )

        prop_val =  get_si_prop( good_Psi_val, targ_prop=prop_desc, ind_name=ind_name, ind_si_val=ind_si_val, symbol=symbol )

    return prop_val, good_Psi_val

if __name__ == "__main__":
    from engcoolprop.ec_fluid import (toEng_callD,  DSI_fromEng ,  
                                      PropsSI ,   SSI_fromEng , TSI_fromEng ,  
                                      UHSI_fromEng )
    from engcoolprop.parameter_units import si_unitsD

    Psi_val = 0 # PSI_fromEng(14.696) # run through all possible calcs at 1 atm
    # print( 'Psi_val =', Psi_val)

    ind_valD = {} # key:T, D, H or S, value: SI value
    ind_valD['T'] = TSI_fromEng( 800.0) # 527.67 )
    ind_valD['D'] =  DSI_fromEng( 55 ) # 62.4273 )
    ind_valD['H'] = UHSI_fromEng( 0.0 )
    ind_valD['S'] = SSI_fromEng( 0.0 )

    for prop_desc in ['T', 'D', 'U', 'H', 'S', 'C', 'V', 'L']:
        print()
        for ind_name in ['T', 'D', 'H', 'S']: # TP, HP, SP, DP
            ind_si_val = ind_valD[ ind_name ]

            prop_val, good_Psi_val = safe_get_INCOMP_prop( prop_desc, Psi_val=Psi_val, ind_name=ind_name, 
                                                          ind_si_val=ind_si_val, symbol='Water', show_warnings=False )

            print( '%s:'%prop_desc, 'at P=%g Pa'%good_Psi_val, 'and %s=%g'%(ind_name, ind_si_val), '(%s)'%si_unitsD[ind_name], 
                   prop_desc + '=' + '%8g'%toEng_callD[prop_desc](prop_val), '(%s)'%si_unitsD[prop_desc] )



import numpy as np


class RootResults:
    def __init__(self, root, iterations, function_calls, flag, converged):
        self.root = root
        self.iterations = iterations
        self.function_calls = function_calls
        self.flag = flag
        self.converged = converged

    def __repr__(self):
        return (f"RootResults(root={self.root}, iterations={self.iterations}, "
                f"function_calls={self.function_calls}, flag={self.flag}, "
                f"converged={self.converged})")


# brentq method take from scipy
# (modified slightly for initial lower and upper bounds within xtol)
def brentq(f, a, b, args=(), xtol=1e-12, rtol=4*np.finfo(float).eps, maxiter=100, **kwargs):
    """
    Brent's method for root finding.
    
    Parameters
    ----------
    f : function
        Python function returning a scalar. The root to be found is such that f(x) = 0.
    a : scalar
        Lower bound of the bracketing interval [a,b].
    b : scalar
        Upper bound of the bracketing interval [a,b].
    xtol : scalar, optional
        Absolute error tolerance. Default is 1e-12.
    rtol : scalar, optional
        Relative error tolerance. Default is 4 * machine epsilon for float.
    maxiter : int, optional
        Maximum number of iterations. Default is 100.
    args : tuple, optional
        Extra arguments for the function `f`.
    
    """
    iters = 0
    funcalls = 1
    fa = f(a, *args)
    if abs(fa) < xtol:
        return RootResults(root=a, iterations=iters, function_calls=funcalls, flag=0, converged=True)
    
    funcalls = 2
    fb = f(b, *args)
    if abs(fb) < xtol:
        return RootResults(root=b, iterations=iters, function_calls=funcalls, flag=0, converged=True)

    if fa * fb > 0:
        print( ' === in brentq, fa=', fa, 'with a=',a)
        print( ' === in brentq, fb=', fb, 'with b=',b)
        raise ValueError("f(a) and f(b) must have different signs")

    c, fc = a, fa
    d = e = b - a

    for iters in range(maxiter):
        if fb * fc > 0:
            c, fc, d, e = a, fa, b - a, b - a

        if abs(fc) < abs(fb):
            a, b, c = b, c, b
            fa, fb, fc = fb, fc, fb

        tol = 2 * xtol * max(abs(b), 1.0) + 0.5 * rtol
        m = 0.5 * (c - b)

        if abs(m) <= tol or fb == 0:
            return RootResults(root=b, iterations=iters, function_calls=funcalls, flag=0, converged=True)

        if abs(e) >= tol and abs(fa) > abs(fb):
            s = fb / fa
            if a == c:
                p = 2 * m * s
                q = 1 - s
            else:
                q = fa / fc
                r = fb / fc
                p = s * (2 * m * q * (q - r) - (b - a) * (r - 1))
                q = (q - 1) * (r - 1) * (s - 1)

            if p > 0:
                q = -q
            p = abs(p)

            if 2 * p < min(3 * m * q - abs(tol * q), abs(e * q)):
                e, d = d, p / q
            else:
                d, e = m, m
        else:
            d, e = m, m

        a, fa = b, fb
        if abs(d) > tol:
            b += d
        else:
            b += np.sign(m) * tol

        fb = f(b, *args)
        funcalls += 1

    return RootResults(root=b, iterations=iters, function_calls=funcalls, flag=1, converged=False)
        

# Example usage
if __name__ == "__main__":

    def root_function(x):
        """
        Define the function for which we want to find the root.
        """
        # Example function: f(x) = x^2 - 2
        return x**2 - 2

    def find_root_brentq(func, a, b, tol=1e-5):
        """
        Find the root of the given function using the Brentq method.
        
        Parameters:
        func (callable): The function for which to find the root.
        a (float): The lower bound of the interval.
        b (float): The upper bound of the interval.
        tol (float): The tolerance for the root-finding algorithm. Default is 1e-5.
        
        Returns:
        float: The root of the function within the given interval.
        """
        sol = brentq(root_function, 0, 2)
        print(sol)
        return sol.root


    a = 0  # Lower bound of the interval
    b = 2  # Upper bound of the interval
    root = find_root_brentq(root_function, a, b)
    print(f"The root of the function in the interval [{a}, {b}] is: {root}")        




        
def floatDammit( val=0.0 ):
    val = str( val )
    v = val.strip()
    if v.lower() in ['nan','inf','-inf']:
        if v.lower() == '-inf':
            v = -1.0E99
        elif v.lower() == 'inf':
            v = 1.0E99
        else:
            v = 0.0
    try:
        v = float(v)
    except:
        v = 0.0
    return v

def is_float( fval ):
    if type(fval)==type(11.11):
        return 1
    if type(fval)==type('string'):
        if fval.find('.')==-1:
            return 0
    try:
        if float(fval) > -float("inf"):
            return 1
    except:
        #print 'Failed is_float',fval,type(fval)
        pass
    return 0

def get_banner_chars( outlineChar=None ):
    
    
    if outlineChar is None:# printable characters
        ul = '+'; ur = '+'; ll = '+'; lr = '+'; vert = '|'; hor = '-'
        vl = vr = vert
        ht = hb = hor
        um = lm = '+'
    else:
        # will be all outlineChar if not printable characters from above
        ul = ur = ll =  lr = vr = vl = hb = ht = vert = hor = um = lm = outlineChar
        
    return ul, ur, ll,  lr, vr, vl, hb, ht, vert, hor, um, lm


def banner(s, outlineChar=None, leftMargin=0, just='left'):
    
    ul, ur, ll,  lr, vr, vl, hb, ht, vert, hor, um, lm = get_banner_chars( outlineChar )
    
    sL = s.split('\n')
    L = 0
    for s in sL:
        ss = '  ' + s + '  '
        L = max(L, len(ss))
    
    top = leftMargin*' ' + ul + L*ht + ur
    bot = leftMargin*' ' + ll + L*hb + lr
    
    midL = []
    for s in sL:
        # if just=='center':
        #     midL.append( leftMargin*' ' + vl + '%s'%(s.center(L),) + vr )
        # elif just=='right':
        #     midL.append( leftMargin*' ' + vl + '%s'%(s.rjust(L),) + vr )
        # else:
        #     midL.append( leftMargin*' ' + vl + '%s'%(s.ljust(L),) + vr )

        midL.append( leftMargin*' ' + vl + '%s'%(s.center(L),) + vr )
        
    
    pad = ''
    if just=='center':
        Lpad = (80 - len(top)) // 2
        pad = ' '*Lpad
    elif just=='right':
        Lpad = 79 - len(top)
        pad = ' '*Lpad

    
    print(pad+top)
    for mid in midL:
        print(pad+mid)
    print(pad+bot)

def getStr( val ):
    if type(val) == type('str'):
        return val
    if is_float(val):
        val = floatDammit( val )
        #return ('%f' % val).rstrip('0').rstrip('.')
        return '%g'%val 
    else:
        return str( val )

def show_table( titleL=None, LOL=None, outlineChar=None, header_sep=True ):
    '''Number of columns = len(titleL)
       Lists of values in list of lists LOL
    '''
    #ul = chr(218); ur = chr(191); ll = chr(192); lr = chr(217); vert = chr(179); hor = chr(196)
    #um = chr(194); lm = chr(193)
    #vl = vr = vert
    #ht = hb = hor
    ul, ur, ll,  lr, vr, vl, hb, ht, vert, hor, um, lm = get_banner_chars( outlineChar )
    
    
    Ncol = len(titleL)
    lenL = [ len(str(t)) for t in titleL ] # allocation length for each column 
    # lenL.append( len(titleL[j]) ) # init to title Length
    for j in range( Ncol ):
        for L in LOL:
            val = L[j]
            lstr = len( getStr(val) )
            if lstr > lenL[j]: # set lenL[j] to max len in column
                lenL[j] = lstr
            

    #print 'lenL =',lenL
    # print top line 
    sL = [ul]
    for i in range(Ncol):
        sL.append( hor*(lenL[i]+2) )
        if i < Ncol-1:
            sL.append( um )
    sL.append(ur)
    print(''.join(sL))
    
    # print titles
    sL = []
    for i,title in enumerate(titleL):
        sL.append( title.center( lenL[i]+2 ) )
    print(vert + vert.join(sL) + vert)

    if header_sep:
        # print header separation line
        sL = [ul]
        for i in range(Ncol):
            sL.append( hor*(lenL[i]+2) )
            if i < Ncol-1:
                sL.append( um )
        sL.append(ur)
        print(''.join(sL))

    # print content
    for i in range( len(LOL) ):
        sL = []
        for j in range( Ncol ):
            sL.append( getStr(LOL[i][j]).center( lenL[j]+2 ) )
        print(vert + vert.join(sL) + vert)

    # print bottom line 
    sL = [ll]
    for i in range(Ncol):
        sL.append( hor*(lenL[i]+2) )
        if i < Ncol-1:
            sL.append( lm )
    sL.append(lr)
    print(''.join(sL))
    
def dev_tests():
        
    for i in range(170, 241, 10):
        for j in range(10):
            print(i+j,chr(i+j),' ', end=' ')
        print()
    banner('Default Banner')
    banner('outlineChar = *\njust = "left"', outlineChar='*')

    banner('outlineChar = None\njust = "center"', outlineChar=None, leftMargin=0, just='center')
    banner('outlineChar = #\njust = "right"', outlineChar='#', leftMargin=0, just='right')
    
    # --------------
    # for i in range(6,8):
    # use either default or i=0 or 7
    for c in [None,  '|', '.']:
        show_table( ['col1 outlineChar=%s'%c,'second Col','col 3'], [[1,2,3.333],['a','seven',123.4]], outlineChar=c )

    print()
    for v in ['nan','inf','-inf', 123, 99.0, 'hi']:
        print( 'v=%-4s'%v,'  getstr=%-4s'%getStr(v), 'floatDammit=',floatDammit(v))
    


if __name__=="__main__":
    dev_tests()    
    



import unittest
# import unittest2 as unittest # for versions of python < 2.7

"""
        Method                            Checks that
self.assertEqual(a, b)                      a == b   
self.assertNotEqual(a, b)                   a != b   
self.assertTrue(x)                          bool(x) is True  
self.assertFalse(x)                         bool(x) is False     
self.assertIs(a, b)                         a is b
self.assertIsNot(a, b)                      a is not b
self.assertIsNone(x)                        x is None 
self.assertIsNotNone(x)                     x is not None 
self.assertIn(a, b)                         a in b
self.assertNotIn(a, b)                      a not in b
self.assertIsInstance(a, b)                 isinstance(a, b)  
self.assertNotIsInstance(a, b)              not isinstance(a, b)  
self.assertAlmostEqual(a, b, places=5)      a within 5 decimal places of b
self.assertNotAlmostEqual(a, b, delta=0.1)  a is not within 0.1 of b
self.assertGreater(a, b)                    a is > b
self.assertGreaterEqual(a, b)               a is >= b
self.assertLess(a, b)                       a is < b
self.assertLessEqual(a, b)                  a is <= b

for expected exceptions, use:

with self.assertRaises(Exception):
    blah...blah...blah

with self.assertRaises(KeyError):
    blah...blah...blah

Test if __name__ == "__main__":


    def test__main__(self):
        # Change bottom of source file to call "dev_tests"
        
         def dev_tests():
            pass

         if __name__ == "__main__":
            dev_tests()
            
        # then test by calling <name>.dev_tests()


See:
      https://docs.python.org/2/library/unittest.html
         or
      https://docs.python.org/dev/library/unittest.html
for more assert options
"""

import sys, os
from CoolProp.CoolProp import PropsSI
from engcoolprop.ec_incomp_soln import EC_Incomp_Soln, dev_tests
from engcoolprop.utils import incomp_pure_solnL

here = os.path.abspath(os.path.dirname(__file__)) # Needed for py.test
up_one = os.path.split( here )[0]  # Needed to find models development version
if here not in sys.path[:2]:
    sys.path.insert(0, here)
if up_one not in sys.path[:2]:
    sys.path.insert(0, up_one)

class MyTest(unittest.TestCase):


    def test_should_always_pass_cleanly(self):
        """Should always pass cleanly."""
        pass

    def test_myclass_existence(self):
        """Check that myclass exists"""

        ec_inc = EC_Incomp_Soln( symbol='ZFC-60%', auto_fix_value_errors=True )
        
        # See if the self.EC_Incomp_Soln object exists
        self.assertIsInstance(ec_inc, EC_Incomp_Soln, msg=None)

    def test_make_all_fluids(self):
        """test make all fluids"""

        for base_symbol in incomp_pure_solnL:
            frac_max = PropsSI('fraction_max','INCOMP::%s'%base_symbol )
            frac_min = PropsSI('fraction_min','INCOMP::%s'%base_symbol )
            pcent = int( 100 * (frac_max + frac_min)/ 2 )
            symbol = base_symbol + '-%i'%pcent + '%'

            ec_inc = EC_Incomp_Soln( symbol=symbol, auto_fix_value_errors=False )
            self.assertIsInstance(ec_inc, EC_Incomp_Soln, msg=None)

    def test_setPropsDP_all_fluids(self):
        """test setPropsDP for all fluids"""

        for base_symbol in incomp_pure_solnL:
            frac_max = PropsSI('fraction_max','INCOMP::%s'%base_symbol )
            frac_min = PropsSI('fraction_min','INCOMP::%s'%base_symbol )
            pcent = int( 100 * (frac_max + frac_min)/ 2 )
            symbol = base_symbol + '-%i'%pcent + '%'

            ec_inc = EC_Incomp_Soln( symbol=symbol, auto_fix_value_errors=False )
            T = ec_inc.T
            ec_inc.setPD( ec_inc.P, ec_inc.D)

            self.assertAlmostEqual(T, ec_inc.T, delta=0.1)
            

    def test_setPropsTP_Tmin_Tmax_all_fluids(self):
        """test setPropsDP for all fluids"""

        for base_symbol in incomp_pure_solnL:
            frac_max = PropsSI('fraction_max','INCOMP::%s'%base_symbol )
            frac_min = PropsSI('fraction_min','INCOMP::%s'%base_symbol )
            pcent = int( 100 * (frac_max + frac_min)/ 2 )
            symbol = base_symbol + '-%i'%pcent + '%'

            ec_inc = EC_Incomp_Soln( symbol=symbol, auto_fix_value_errors=True )
            
            # check that safeguards for Tmin/Tmax are working
            ec_inc.setTP( ec_inc.Tmin-1, ec_inc.P)
            ec_inc.setTP( ec_inc.Tmax+1, ec_inc.P)

            

    def test__main__(self):
        old_sys_argv = list(sys.argv)
        sys.argv = list(sys.argv)
        sys.argv.append('suppress_show')
        
        try:
            dev_tests()
        except:
            raise Exception('ERROR... failed in __main__ routine')
        finally:
            sys.argv = old_sys_argv
        

if __name__ == '__main__':
    # Can test just this file from command prompt
    #  or it can be part of test discovery from nose, unittest, pytest, etc.
    unittest.main()




import CoolProp.CoolProp as CP
from CoolProp import __fluids__

def list_heos_fluids():
    heos_fluids = []
    for fluid in CP.FluidsList():
        try:
            CP.AbstractState("HEOS", fluid)
            heos_fluids.append(fluid)
        except ValueError:
            continue
    return heos_fluids

if __name__ == "__main__":

    # List all fluids in HEOS backend
    heos_fluids = list_heos_fluids()
    print("Fluids available in HEOS backend:")
    print( heos_fluids )

    print( '.'*66 )
    print( '__fluids__ =', __fluids__)




from engcoolprop.ec_fluid import EC_Fluid

# Print state point
ec = EC_Fluid(symbol="N2", T=530.0,P=100.0 ) # T=degR, P=psia
ec.setProps(T=500., D=0.1)

ec.printProps() # Print state point at given T,P

import engcoolprop.ec_fluid
pL = dir( engcoolprop.ec_fluid )

print( 'from engcoolprop.ec_fluid import (', end=' ')         
for p in pL:
    if not p.startswith( '__'):
     print( p, ', ', end=' ')
print( ')' )
print()

print( engcoolprop.ec_fluid.call_tuplesD )



""" 
Humid Air Properties
http://www.coolprop.org/fluid_properties/HumidAir.html#table-of-inputs-outputs-to-hapropssi

"""


class EC_Humid_Air(object):
    
    def __init__(self, T=None ,P=None):
        pass



import CoolProp.CoolProp as CP
from CoolProp import __incompressibles_pure__, __incompressibles_solution__

def list_incompressible_fluids():
    # Retrieve the list of all available fluids
    incompressible_fluids = CP.get_global_param_string('incompressible_list_pure').split(',')
    return incompressible_fluids

def list_incompressible_solutions():
    # Retrieve the list of all available fluids
    incompressible_solutions = CP.get_global_param_string('incompressible_list_solution').split(',')
    return incompressible_solutions

if __name__ == "__main__":
    fluids = list_incompressible_fluids()
    print( fluids )

    solutions = list_incompressible_solutions()
    print( solutions )

    print( '.'*66 )
    print( '__incompressibles_pure__ =', __incompressibles_pure__)

    print( '.'*66 )
    print( '__incompressibles_solution__ =', __incompressibles_solution__)



from engcoolprop.ec_fluid import EC_Fluid

ec = EC_Fluid(symbol="AR")
print('Properties for: %s (%s)'%(ec.name, ec.symbol) )
for T in range(460, 570, 10):
    ec.setProps(T=T, P=14.7)
    ec.printTPD()




import CoolProp.CoolProp as CP

# Example fluid: Water
fluid = 'INCOMP::Water'

# Specifying Temperature and Pressure
T = 298.15  # Temperature in K
P = 101325  # Pressure in Pa

density = CP.PropsSI('D', 'T', T, 'P', P, fluid)
specific_heat_capacity = CP.PropsSI('C', 'T', T, 'P', P, fluid)
internal_energy = CP.PropsSI('U', 'T', T, 'P', P, fluid)
enthalpy = CP.PropsSI('H', 'T', T, 'P', P, fluid)
entropy = CP.PropsSI('S', 'T', T, 'P', P, fluid)
viscosity = CP.PropsSI('V', 'T', T, 'P', P, fluid)
thermal_conductivity = CP.PropsSI('L', 'T', T, 'P', P, fluid)

propertyD = {} # key:property letter (e.g. D, P, T), value:numeric SI value
propertyD['D'] = density
propertyD['C'] = specific_heat_capacity
propertyD['U'] = internal_energy
propertyD['H'] = enthalpy
propertyD['S'] = entropy
propertyD['V'] = viscosity
propertyD['L'] = thermal_conductivity
propertyD['T'] = T
propertyD['P'] = P


all_propL = sorted( propertyD.keys() )

call_tuplesD = {} # key:tuple indep vars (e.g. ("T","P")), value: set of dependent properties (e.g. "S", "H", etc.)

num_workedD = {} # key:independent property pair, value:number of successful calcs
# Iterate through all pairs of properties
for prop1 in all_propL:
    for prop2 in all_propL:
        if prop1 != prop2:
            for prop in all_propL:
                if prop not in [prop1, prop2]:
                    try:
                        val = CP.PropsSI( prop, prop1, propertyD[prop1], prop2, propertyD[prop2], fluid)
                        print( 'success: %s%s for property:%s = %g'%(prop1, prop2, prop, val))
                        
                        t = (prop1, prop2)
                        if t in call_tuplesD:
                            call_tuplesD[t].add( prop )
                        else:
                            call_tuplesD[t] = set( [prop] )

                        if prop2 > prop1:
                            pair = prop1 + prop2
                        else:
                            pair = prop2 + prop1

                        num_workedD[pair] = num_workedD.get( pair, 0) + 1
                    except:
                        print( 'FAILED: %s%s for property:%s'%(prop1, prop2, prop))

print( num_workedD )
for k,v in num_workedD.items():
    print( k, v)

"""
Temperature (T)
Pressure (P)
Density (D)
Specific Heat Capacity (C)
Internal Energy (U)
Enthalpy (H)
Entropy (S)
Viscosity (V)
Thermal Conductivity (L)
"""
print( 'call_tuplesD = {}  # key:tuple indep vars (e.g. ("T","P")), value: dependent property (e.g. "S", "H", etc.)')
for t,v in call_tuplesD.items():
    print( 'call_tuplesD[%s] = %s'%(t,v))



import unittest
# import unittest2 as unittest # for versions of python < 2.7

"""
        Method                  Checks that
self.assertEqual(a, b)           a == b   
self.assertNotEqual(a, b)        a != b   
self.assertTrue(x)               bool(x) is True  
self.assertFalse(x)              bool(x) is False     
self.assertIs(a, b)              a is b
self.assertIsNot(a, b)           a is not b
self.assertIsNone(x)             x is None 
self.assertIsNotNone(x)          x is not None 
self.assertIn(a, b)              a in b
self.assertNotIn(a, b)           a not in b
self.assertIsInstance(a, b)      isinstance(a, b)  
self.assertNotIsInstance(a, b)   not isinstance(a, b)  

See:
      https://docs.python.org/2/library/unittest.html
         or
      https://docs.python.org/dev/library/unittest.html
for more assert options
"""

import sys, os

here = os.path.abspath(os.path.dirname(__file__)) # Needed for py.test
up_one = os.path.split( here )[0]  # Needed to find engcoolprop development version
if here not in sys.path[:2]:
    sys.path.insert(0, here)
if up_one not in sys.path[:2]:
    sys.path.insert(0, up_one)

from engcoolprop.ec_fluid import EC_Fluid

class MyTest(unittest.TestCase):


    def test_should_always_pass_cleanly(self):
        """Should always pass cleanly."""
        self.assertAlmostEqual(0.0, 0.0000005, places=5)
        pass

    def test_myclass_existence(self):
        """Check that myclass exists"""
        ec = EC_Fluid( symbol="N2" )

        # See if the self.myclass object exists
        self.assertTrue( ec )
        del( ec )

    def test_setProps_N2(self):
        '''Check call to setProps with fluid N2'''
        ec = EC_Fluid( symbol="N2", T=200., P=300. )
        ec.setProps( T=530.0, P=1000.0)
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_setProps_O2(self):
        '''Check call to setProps with fluid O2'''
        ec = EC_Fluid( symbol="O2" , T=200., P=300.)
        ec.setProps( T=530.0, P=1000.0)
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_dH_FromHref_N2(self):
        '''Check call to dH_FromHref with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        dH = ec.dH_FromHref()
        
        self.assertAlmostEqual(dH, 177.5413109961724 , places=3)

        del( ec )


    def test_dH_FromHref_O2(self):
        '''Check call to dH_FromHref with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        dH = ec.dH_FromHref()
        
        self.assertAlmostEqual(dH, 165.11167057591305, places=3)

        del( ec )


    def test_setPropsFromAS_N2(self):
        '''Check call to setPropsFromAS with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.setPropsFromAS()
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_setPropsFromAS_O2(self):
        '''Check call to setPropsFromAS with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.setPropsFromAS()
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_setTP_N2(self):
        '''Check call to setTP with fluid N2'''
        ec = EC_Fluid( symbol="N2", T=200., P=300. )
        ec.setTP(T=530.0,P=1000.0)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_setTP_O2(self):
        '''Check call to setTP with fluid O2'''
        ec = EC_Fluid( symbol="O2", T=200., P=300. )
        ec.setTP(T=530.0,P=1000.0)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_newDE_N2(self):
        '''Check call to newDE with fluid N2'''
        ec = EC_Fluid( symbol="N2" , T=200., P=300.)
        
        ec.newDE(D=4.94659381374,E=87.6731930702)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=2)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_newDE_O2(self):
        '''Check call to newDE with fluid O2'''
        ec = EC_Fluid( symbol="O2" , T=200., P=300.)
        ec.newDE(D=5.86052712206,E=76.1981078385)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=2)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_constP_newH_N2(self):
        '''Check call to constP_newH with fluid N2'''
        ec = EC_Fluid( symbol="N2" , T=200., P=1000.)
        ec.constP_newH(125.082832491)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_constP_newH_O2(self):
        '''Check call to constP_newH with fluid O2'''
        ec = EC_Fluid( symbol="O2" , T=200., P=1000.)
        ec.constP_newH(107.773815464 )
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=4)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_constH_newP_N2(self):
        '''Check call to constH_newP with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.setProps(P=500., H=125.082832491)
        ec.constH_newP(P=1000.0)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_constH_newP_O2(self):
        '''Check call to constH_newP with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.setProps(P=500., H=107.773815464)
        ec.constH_newP(P=1000.0)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=4)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_setPH_N2(self):
        '''Check call to setPH with fluid N2'''
        ec = EC_Fluid( symbol="N2", T=200., P=1000. )
        ec.setPH(1000.0,125.082832491)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_setPH_O2(self):
        '''Check call to setPH with fluid O2'''
        ec = EC_Fluid( symbol="O2" , T=200., P=1000.)
        ec.setPH(1000.0,107.773815464)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 5.86052182537, places=4)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_constS_newP_N2(self):
        '''Check call to constS_newP with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.setProps(T=500., S=1.31927325217)
        ec.constS_newP(P=1000.0)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_constS_newP_O2(self):
        '''Check call to constS_newP with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.setProps(T=500., S=1.25461825929)
        ec.constS_newP(P=1000.0)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_setTD_N2(self):
        '''Check call to setTD with fluid N2'''
        ec = EC_Fluid( symbol="N2" , T=200., P=300.)
        ec.setTD(T=530.0,D=4.94659381374)
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_setTD_O2(self):
        '''Check call to setTD with fluid O2'''
        ec = EC_Fluid( symbol="O2" , T=200., P=300. )
        ec.setTD(T=530.0,D=5.86052712206)
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_setPD_N2(self):
        '''Check call to setPD with fluid N2'''
        ec = EC_Fluid( symbol="N2"  , T=200., P=300.)
        ec.setPD(P=1000.0,D=4.94659381374)
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_setPD_O2(self):
        '''Check call to setPD with fluid O2'''
        ec = EC_Fluid( symbol="O2"  , T=200., P=300.)
        ec.setPD(P=1000.0,D=5.86052712206)
        
        self.assertAlmostEqual(ec.T, 530.0, places=3)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_getSatP_N2(self):
        '''Check call to getSatP with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        Psat = ec.getSatP( T=None)
        Psat = ec.getSatP( T=200.0)
        
        self.assertAlmostEqual(Psat, 226.61149712125732, places=5)

        del( ec )


    def test_getSatP_O2(self):
        '''Check call to getSatP with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        Psat = ec.getSatP( T=None)
        Psat = ec.getSatP( T=200.0)
        
        self.assertAlmostEqual(Psat, 85.02309450018438, places=5)


        del( ec )


    def test_getSatPandDens_N2(self):
        '''Check call to getSatPandDens with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        Psat, Dliq, Dgas = ec.getSatPandDens( T=None)
        Psat, Dliq, Dgas = ec.getSatPandDens( T=200.0)
        
        self.assertAlmostEqual(Psat, 226.61149712125732, places=5)
        self.assertAlmostEqual(Dliq, 38.2470012845856, places=5)
        self.assertAlmostEqual(Dgas, 4.946593813736714, places=5)

        del( ec )


    def test_getSatPandDens_O2(self):
        '''Check call to getSatPandDens with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        Psat, Dliq, Dgas = ec.getSatPandDens( T=None)
        Psat, Dliq, Dgas = ec.getSatPandDens( T=200.0)
        
        self.assertAlmostEqual(Psat, 85.02309450018438, places=5)
        self.assertAlmostEqual(Dliq, 64.23715727948591, places=5)
        self.assertAlmostEqual(Dgas, 5.860527122060697, places=5)

        del( ec )


    def test_getSatT_N2(self):
        '''Check call to getSatT with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        Tsat = ec.getSatT( P=None)
        Tsat = ec.getSatT( P=80.0)
        
        self.assertAlmostEqual(Tsat, 171.4764377703347, places=5)

        del( ec )


    def test_getSatT_O2(self):
        '''Check call to getSatT with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.getSatT( P=None)
        Tsat = ec.getSatT( P=None)
        Tsat = ec.getSatT( P=80.0)
        
        self.assertAlmostEqual(Tsat, 198.39085041700642, places=5)

        del( ec )


    def test_getSatTandDens_N2(self):
        '''Check call to getSatTandDens with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        Tsat, Dliq, Dgas = ec.getSatTandDens(P=None)
        Tsat, Dliq, Dgas = ec.getSatTandDens(P=80.0)
        
        self.assertAlmostEqual(Tsat, 171.4764377703347, places=5)
        self.assertAlmostEqual(Dliq, 44.747923887617794, places=5)
        self.assertAlmostEqual(Dgas, 1.4181440790822393, places=5)

        del( ec )


    def test_getSatTandDens_O2(self):
        '''Check call to getSatTandDens with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.getSatTandDens(P=None)
        Tsat, Dliq, Dgas = ec.getSatTandDens(P=80.0)
        
        self.assertAlmostEqual(Tsat, 198.39085041700642, places=5)
        self.assertAlmostEqual(Dliq, 64.56368984711504, places=5)
        self.assertAlmostEqual(Dgas, 1.3475849732116867, places=5)

        del( ec )


    def test_dHvap_N2(self):
        '''Check call to dHvap with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        dH = ec.dHvap( P=None)
        dH = ec.dHvap( P=15.0)
        
        self.assertAlmostEqual(dH, 85.53357151545947, places=4)

        del( ec )


    def test_dHvap_O2(self):
        '''Check call to dHvap with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        dH = ec.dHvap( P=None)
        dH = ec.dHvap( P=15.0)
        
        self.assertAlmostEqual(dH, 91.51502631501589, places=4)

        del( ec )


    def test_surfTen_N2(self):
        '''Check call to surfTen with fluid N2'''
        ec = EC_Fluid( symbol="N2", P=80, T=200 )
        ec.setProps(T=200, Q=0)
        st = ec.surfTen()        
        self.assertAlmostEqual(st, 1.1726812809126956e-05, places=5)

        del( ec )


    def test_surfTen_O2(self):
        '''Check call to surfTen with fluid O2'''
        ec = EC_Fluid( symbol="O2", P=80, T=200 )
        ec.setProps(T=200, Q=0)
        st = ec.surfTen()
        self.assertAlmostEqual(st, 4.638551601014212e-05, places=5)

        del( ec )


    def test_restoreFromDup_N2(self):
        '''Check call to restoreFromDup with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.restoreFromDup()
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_restoreFromDup_O2(self):
        '''Check call to restoreFromDup with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.restoreFromDup()
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=3)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_saveToDup_N2(self):
        '''Check call to saveToDup with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.saveToDup()
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_saveToDup_O2(self):
        '''Check call to saveToDup with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.saveToDup()
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_initFromObj_N2(self):
        '''Check call to initFromObj with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.initFromObj( ec )
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 4.94659381374, places=5)
        self.assertAlmostEqual(ec.S, 1.31927325217, places=5)

        del( ec )


    def test_initFromObj_O2(self):
        '''Check call to initFromObj with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.initFromObj( ec )
        
        self.assertAlmostEqual(ec.T, 530.0, places=5)
        self.assertAlmostEqual(ec.P, 1000.0, places=5)
        self.assertAlmostEqual(ec.D, 5.86052712206, places=5)
        self.assertAlmostEqual(ec.S, 1.25461825929, places=5)

        del( ec )


    def test_gamma_N2(self):
        '''Check call to gamma with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        gam = ec.gamma()
        
        self.assertAlmostEqual(gam, 1.5214910882792643, places=5)

        del( ec )


    def test_gamma_O2(self):
        '''Check call to gamma with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        gam = ec.gamma()
        
        self.assertAlmostEqual(gam, 1.539276126838285, places=5)

        del( ec )


    def test_getStrTransport_N2(self):
        '''Check call to getStrTransport with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.getStrTransport()
        
        del( ec )


    def test_getStrTransport_O2(self):
        '''Check call to getStrTransport with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.getStrTransport()
        
        del( ec )


    def test_getStrTPD_N2(self):
        '''Check call to getStrTPD with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.getStrTPD()
        
        del( ec )


    def test_getStrTPD_O2(self):
        '''Check call to getStrTPD with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.getStrTPD()
        
        del( ec )


    def test_getStrTPDphase_N2(self):
        '''Check call to getStrTPDphase with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.getStrTPDphase()

        del( ec )


    def test_getStrTPDphase_O2(self):
        '''Check call to getStrTPDphase with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.getStrTPDphase()

        del( ec )


    def test_printTPD_N2(self):
        '''Check call to printTPD with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.printTPD()

        del( ec )


    def test_printTPD_O2(self):
        '''Check call to printTPD with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.printTPD()

        del( ec )


    def test_Qdescription_N2(self):
        '''Check call to Qdescription with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        s = ec.Qdescription()
        
        self.assertEqual(s, 'Supercritical Fluid')

        del( ec )


    def test_Qdescription_O2(self):
        '''Check call to Qdescription with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        s = ec.Qdescription()
        
        self.assertEqual(s, 'Supercritical Fluid')

        del( ec )


    def test_html_desc_N2(self):
        '''Check call to html_desc with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.html_desc()

        del( ec )


    def test_html_desc_O2(self):
        '''Check call to html_desc with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.html_desc()

        del( ec )


    def test_printCriticalProps_N2(self):
        '''Check call to printCriticalProps with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.printCriticalProps()

        del( ec )


    def test_printCriticalProps_O2(self):
        '''Check call to printCriticalProps with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.printCriticalProps()

        del( ec )


    def test_printProps_N2(self):
        '''Check call to printProps with fluid N2'''
        ec = EC_Fluid( symbol="N2" )
        ec.printProps()

        del( ec )


    def test_printProps_O2(self):
        '''Check call to printProps with fluid O2'''
        ec = EC_Fluid( symbol="O2" )
        ec.printProps()

        del( ec )


if __name__ == '__main__':
    # Can test just this file from command prompt
    #  or it can be part of test discovery from nose, unittest, pytest, etc.
    unittest.main()




import CoolProp.CoolProp as CP

def is_heos_fluid(fluid_name):
    try:
        CP.AbstractState("HEOS", fluid_name)
        return True
    except ValueError:
        return False

# Example usage
for fluid_name in ['n2', 'N2', 'XXX', 'DOWQ']:
    if is_heos_fluid(fluid_name):
        print(f"{fluid_name} is available in HEOS fluids.")
    else:
        print(f"{fluid_name} is not available in HEOS fluids.")

print( '='*66)


def is_incomp_fluid(fluid_name):
    try:
        CP.AbstractState("INCOMP", fluid_name)
        return True
    except ValueError:
        return False

# Example usage
for fluid_name in [ 'DOWQ', 'DowQ']:
    if is_incomp_fluid(fluid_name):
        print(f"{fluid_name} is available in INCOMP fluids.")
    else:
        print(f"{fluid_name} is not available in INCOMP fluids.")



from engcoolprop.ec_incomp_soln import EC_Incomp_Soln

# Create incompressible solution object. (without specifying state point)
ec_soln = EC_Incomp_Soln(symbol="MEG-30%")

# Print state point
ec_soln.printProps() # Print state point

print( '\n\n\n\n' )

ec_soln.printTPD()

ec_soln.printTransport()



import CoolProp.CoolProp as CP
from CoolProp.Plots import PropertyPlot
from CoolProp.Plots.Common import BasePlot, PropertyDict, BaseDimension
from engcoolprop.ec_fluid import EC_Fluid

symbol = 'N2'

class ENGunits(PropertyDict):
    def __init__(self):
        self._D = BaseDimension(add_SI=0.0, mul_SI=1./16.01843417, off_SI=0.0, label='Density',       symbol=u'd', unit=u'lbm/ft3')
        self._H = BaseDimension(add_SI=0.0, mul_SI=0.000429923, off_SI=0.0, label='Specific Enthalpy',symbol=u'h', unit=u'BTU/lbm')
        self._P = BaseDimension(add_SI=0.0, mul_SI=1.0/6894.76, off_SI=0.0, label='Pressure',         symbol=u'p', unit=u'psia')
        self._S = BaseDimension(add_SI=0.0, mul_SI=0.0002388461,off_SI=0.0, label='Specific Entropy', symbol=u's', unit=u'BTU/lbm/R')
        self._T = BaseDimension(add_SI=0.0, mul_SI=1.8,         off_SI=0.0, label='Temperature',      symbol=u'T', unit=u'R')
        self._U = BaseDimension(add_SI=0.0, mul_SI=0.000429923, off_SI=0.0, label='Specific Internal Energy', symbol=u'u', unit=u'BTU/lbm')
        self._Q = BaseDimension(add_SI=0.0, mul_SI=1.0,         off_SI=0.0, label='Vapour Quality',   symbol=u'x', unit=u'')
        
BasePlot.UNIT_SYSTEMS['ENG'] = ENGunits() # ['EUR','KSI','SI']

ec = EC_Fluid(symbol)
ec.printCriticalProps()

# Plot Types: 'PS', 'PT', 'HS', 'TS', 'PD', 'TD', 'PH'
# tp_limits:  'NONE', 'DEF','ACHP','ORC'
plot = PropertyPlot(symbol, 'TD', unit_system='ENG', tp_limits='DEF')
plot.calc_isolines(CP.iHmass, num=11)
plot.calc_isolines(CP.iQ, num=11)
plot.title('%s (%s)'%(ec.name, ec.symbol))
plot.grid()
plot.show()




from CoolProp.CoolProp import PropsSI

#Specific heat capacity of Downtherm Q at 500 K and 1 atm
print( ' '*18, PropsSI('C','T',500,'P',101325,'INCOMP::DowQ') )
print( 'expected: Out[2]: 2288.1643758645673 ' )

#Density of Downtherm Q at 500 K and 1 atm.
print( ' '*18, PropsSI('D','T',500,'P',101325,'INCOMP::DowQ') )
print( 'expected: Out[3]: 809.0654931667821' )

#Round trip in thermodynamic properties
T_init = 500.0

P_init = 101325

D_init = PropsSI('D','T',T_init,'P',P_init,'INCOMP::DowQ')

S_init = PropsSI('S','D',D_init,'P',P_init,'INCOMP::DowQ')

H_init = PropsSI('H','S',S_init,'P',P_init,'INCOMP::DowQ')

T_init = PropsSI('T','H',H_init,'P',P_init,'INCOMP::DowQ')

print( ' '*18, 'T_init =', T_init )
print( 'expected: Out[10]: 500.0000000000001' )

#Saturation pressure of Downtherm Q at 500 K
print( ' '*18, PropsSI('P','T',500,'Q',0,'INCOMP::DowQ') )
print( 'expected: Out[11]: 38091.37403658103' )

#Minimum temperature for Downtherm Q
print( ' '*18, PropsSI('Tmin','T',0,'P',0,'INCOMP::DowQ') )
print( 'expected: Out[12]: 238.15' )

#Maximum temperature for Downtherm Q
print( ' '*18, PropsSI('Tmax','T',0,'P',0,'INCOMP::DowQ') )
print( 'expected: Out[13]: 633.15' )



from engcoolprop.ec_incomp_fluid import EC_Incomp_Fluid
from engcoolprop.ec_fluid import ( Peng_fromSI, PropsSI , TSI_fromEng )

# Print state point
ec_inc = EC_Incomp_Fluid(symbol="DowQ", T=630.0,P=100.0 ) # T=degR, P=psia

# print( "NOTE: This is EXPECTED TO FAIL!!!")
# ec_inc.setProps(T=500., D=0.1)

ec_inc.printProps() # Print state point at given T,P

t_range = (ec_inc.Tmax - ec_inc.Tmin) / 1.1
N = 9
tL = [ec_inc.Tmin + (i/N)*t_range for i in range(N)]

for T in tL:
    Psat = Peng_fromSI( PropsSI('P','T', TSI_fromEng( T ) ,'Q',0, ec_inc.fluid) )
    print( 'T = %g'%T, 'Psat = %f'%Psat)

Psat_max = Peng_fromSI( PropsSI('P','T', TSI_fromEng( ec_inc.Tmax ) ,'Q',0, ec_inc.fluid) )
print( 'Tmax = %g'%ec_inc.Tmax, 'Psat_max = %g'%Psat_max)



import traceback


def find_exception_limit(f, tolerance=1e-6, lower_bound=0, upper_bound=10, 
                         no_exception='lower_bound', show_warnings=True):
    """
    Finds the value of x where f(x) does NOT throw an exception within the given tolerance.

    Parameters:
    f (function): The function to test.
    tolerance (float): The tolerance within which to find the value of x.
    lower_bound (float): The lower bound of the search interval.
    upper_bound (float): The upper bound of the search interval.

    no_exception(string): default="lower_bound": If no exceptions found return either lower_bound or upper_bound

    Returns:
    float: The value of x where f(x) does NOT throw an exception
    """
    tb_str = ''
    try:
        f( lower_bound )
        bad_lower = False
    except:
        bad_lower = True
        tb_str = traceback.format_exc()
        
    try:
        f( upper_bound )
        bad_upper = False
    except:
        bad_upper = True
        tb_str += '\n\n' + traceback.format_exc()
        
    # Exception boundary not included in bounds
    if bad_lower and bad_upper:
        if show_warnings:
            print( '*** Both upper and lower limits throw Exception in find_exception_limit***' )
            print( '    TRACEBACK:', tb_str)
        return None
        
    # All good
    if not bad_lower and not bad_upper:
        # print( '***No Exception found in find_exception_limit***' )

        # return either lower_bound or upper_bound if no exception found
        if no_exception=='lower_bound':
            return lower_bound
        else:
            return upper_bound
        
        
    while upper_bound - lower_bound > tolerance:
        mid = (lower_bound + upper_bound) / 2
        try:
            f(mid)
            if bad_lower:
                upper_bound = mid
            else:
                lower_bound = mid
        except Exception:
            if bad_lower:
                lower_bound = mid
            else:
                upper_bound = mid

    # Check and return the bound that does NOT throw an exception
    if bad_lower:
        return upper_bound
    else:
        return lower_bound
    #try:
    #    f(lower_bound)
    #    return lower_bound
    #except Exception:
    #    return upper_bound

if __name__ == "__main__":
    def f(x):
        # Dummy implementation of the function that may throw an exception.
        # Replace this with the actual function implementation as needed.
        if x >= 5:
            raise ValueError("x is too large")
        return x

    tolerance = 1e-6
    exception_value = find_exception_limit(f, tolerance, show_warnings=True)
    print(f"The value of x where f(x) throws an exception is approximately: {exception_value}")
